@@ -1,19 +1,21 @@
 """Collection of utilities to detect properties of the underlying architecture."""
 
-from subprocess import PIPE, Popen, DEVNULL
+from subprocess import PIPE, Popen, DEVNULL, run
 
 from cached_property import cached_property
 import cpuinfo
+import ctypes
 import numpy as np
 import psutil
 import re
-import ctypes
+import os
+import sys
 
 from devito.logger import warning
 from devito.tools import as_tuple, all_equal, memoized_func
 
 __all__ = ['platform_registry', 'get_cpu_info', 'get_gpu_info', 'get_nvidia_cc',
-           'check_cuda_runtime',
+           'get_cuda_path', 'check_cuda_runtime', 'get_m1_llvm_path',
            'Platform', 'Cpu64', 'Intel64', 'Amd', 'Arm', 'Power', 'Device',
            'NvidiaDevice', 'AmdDevice', 'IntelDevice',
            'INTEL64', 'SNB', 'IVB', 'HSW', 'BDW', 'SKX', 'KNL', 'KNL7210',  # Intel
@@ -220,7 +222,7 @@ def get_gpu_info():
 
                     try:
                         line = lines.pop(deviceid)
-                        _, v, unit = re.split('([0-9]+)\s', line)
+                        _, v, unit = re.split(r'([0-9]+)\s', line)
                         assert unit == 'MiB'
                         return int(v)*10**6
                     except:
@@ -355,6 +357,58 @@ def get_nvidia_cc():
         return 10*cc_major.value + cc_minor.value
 
 
+@memoized_func
+def get_cuda_path():
+    # *** First try: via commonly used environment variables
+    for i in ['CUDA_HOME', 'CUDA_ROOT']:
+        cuda_home = os.environ.get(i)
+        if cuda_home:
+            return cuda_home
+
+    # *** Second try: inspect the LD_LIBRARY_PATH
+    llp = os.environ.get('LD_LIBRARY_PATH', '')
+    for i in llp.split(':'):
+        if re.match('.*/nvidia/hpc_sdk/.*/compilers/lib', i):
+            cuda_home = os.path.join(os.path.dirname(os.path.dirname(i)), 'cuda')
+            # Sanity check
+            if os.path.exists(cuda_home):
+                return cuda_home
+
+    return None
+
+
+@memoized_func
+def get_m1_llvm_path(language):
+    # Check if Apple's llvm is installed (installable via Homebrew), which supports
+    # OpenMP.
+    # Check that we are on apple system
+    if sys.platform != 'darwin':
+        raise ValueError('Apple LLVM is only available on Mac OS X.')
+    # *** First check if LLVM is installed
+    ver = run(["clang", "--version"], stdout=PIPE, stderr=DEVNULL).stdout.decode("utf-8")
+    # *** Second check if this clang version targets arm64 (M1)
+    if "arm64-apple" in ver:
+        # *** Third extract install path. clang version command contains the line:
+        # InstalledDir: /path/to/llvm/bin
+        # which points to the llvm root directory we need for libraries and includes
+        prefix = [v.split(': ')[-1] for v in ver.split('\n')
+                  if "InstalledDir" in v][0][:-4]
+        # ** Fourth check if the libraries are installed
+        if os.path.exists(os.path.join(prefix, 'lib', 'libomp.dylib')):
+            libs = os.path.join(prefix, "lib")
+            include = os.path.join(prefix, "include")
+            return {'libs': libs, 'include': include}
+        elif language == "openmp":
+            warning("Apple's arm64 clang found but openmp libraries not found."
+                    "Install Apple LLVM for OpenMP support, i.e. `brew install llvm`")
+        else:
+            pass
+    else:
+        if language == "openmp":
+            warning("Apple's x86 clang found, OpenMP is not supported.")
+    return None
+
+
 @memoized_func
 def check_cuda_runtime():
     libnames = ('libcudart.so', 'libcudart.dylib', 'cudart.dll')
@@ -367,6 +421,7 @@ def check_cuda_runtime():
             break
     else:
         warning("Unable to check compatibility of NVidia driver and runtime")
+        return
 
     driver_version = ctypes.c_int()
     runtime_version = ctypes.c_int()
