@@ -104,7 +104,7 @@ class Data(np.ndarray):
         # that only one object (the "root" Data) will free the C-allocated memory
         self._memfree_args = None
 
-        if type(obj) != Data:
+        if not issubclass(type(obj), Data):
             # Definitely from view casting
             self._is_distributed = False
             self._modulo = tuple(False for i in range(self.ndim))
@@ -247,7 +247,14 @@ class Data(np.ndarray):
                 else:
                     pass
                 it.iternext()
-            return retval
+            # Check if dimensions of the view should now be reduced to
+            # be consistent with those of an equivalent NumPy serial view
+            reshape = tuple([s for s, i in zip(retval.shape, loc_idx)
+                             if type(i) is not np.int64])
+            if reshape and (0 not in reshape) and (reshape != retval.shape):
+                return retval.reshape(reshape)
+            else:
+                return retval
         elif loc_idx is NONLOCAL:
             # Caller expects a scalar. However, `glb_idx` doesn't belong to
             # self's data partition, so None is returned
