@@ -4,11 +4,9 @@ from collections.abc import Iterable
 from cached_property import cached_property
 import sympy
 
-from devito.finite_differences.finite_difference import (generic_derivative,
-                                                         first_derivative,
-                                                         cross_derivative)
-from devito.finite_differences.differentiable import Differentiable
-from devito.finite_differences.tools import direct, transpose
+from .finite_difference import generic_derivative, first_derivative, cross_derivative
+from .differentiable import Differentiable
+from .tools import direct, transpose
 from devito.tools import as_mapper, as_tuple, filter_ordered, frozendict
 from devito.types.utils import DimensionTuple
 
@@ -203,14 +201,18 @@ class Derivative(sympy.Derivative, Differentiable):
         _kwargs.update(**kwargs)
         return Derivative(expr, *self.dims, **_kwargs)
 
-    @property
-    def func(self):
-        return lambda *a, **kw: self._new_from_self(expr=a[0], **kw)
+    def func(self, expr, *args, **kwargs):
+        return self._new_from_self(expr=expr, **kwargs)
 
     def subs(self, *args, **kwargs):
         """
         Bypass sympy.Subs as Devito has its own lazy evaluation mechanism.
         """
+        # Check if we are calling subs(self, old, new, **hint) in which case
+        # return the standard substitution. Need to check `==` rather than `is`
+        # because a new derivative could be created i.e `f.dx.subs(f.dx, y)`
+        if len(args) == 2 and args[0] == self:
+            return args[1]
         try:
             rules = dict(*args)
         except TypeError:
@@ -223,6 +225,13 @@ class Derivative(sympy.Derivative, Differentiable):
         This is a helper method used internally by SymPy. We exploit it to postpone
         substitutions until evaluation.
         """
+        # Check if trying to replace the whole expression
+        if self in subs:
+            new = subs.pop(self)
+            try:
+                return new._xreplace(subs)
+            except AttributeError:
+                return new, True
         subs = self._ppsubs + (subs,)  # Postponed substitutions
         return self._new_from_self(subs=subs), True
 
@@ -311,48 +320,52 @@ class Derivative(sympy.Derivative, Differentiable):
             # the expression as is.
             return self._new_from_self(x0=x0)
 
-    @property
-    def evaluate(self):
+    def _evaluate(self, **kwargs):
         # Evaluate finite-difference.
-        # Note: evaluate and _eval_fd splitted for potential future different
-        # types of discretizations.
-        return self._eval_fd(self.expr)
+        # NOTE: `evaluate` and `_eval_fd` split for potential future different
+        # types of discretizations
+        return self._eval_fd(self.expr, **kwargs)
 
     @property
     def _eval_deriv(self):
         return self._eval_fd(self.expr)
 
-    def _eval_fd(self, expr):
+    def _eval_fd(self, expr, **kwargs):
         """
-        Evaluate finite difference approximation of the Derivative.
-        Evaluation is carried out via the following four steps:
+        Evaluate the finite-difference approximation of the Derivative.
+        Evaluation is carried out via the following three steps:
 
         - 1: Evaluate derivatives within the expression. For example given
             `f.dx * g`, `f.dx` will be evaluated first.
         - 2: Evaluate the finite difference for the (new) expression.
-        - 3: Evaluate remaining terms (as `g` may need to be evaluated
-        at a different point).
-        - 4: Apply substitutions.
+             This in turn is a two-step procedure, for Functions that may
+             may need to be evaluated at a different point due to e.g. a
+             shited derivative.
+        - 3: Apply substitutions.
         """
         # Step 1: Evaluate derivatives within expression
-        expr = getattr(expr, '_eval_deriv', expr)
+        try:
+            expr = expr._evaluate(**kwargs)
+        except AttributeError:
+            pass
+
+        # If True, the derivative will be fully expanded as a sum of products,
+        # otherwise an IndexSum will returned
+        expand = kwargs.get('expand', True)
 
         # Step 2: Evaluate FD of the new expression
         if self.side is not None and self.deriv_order == 1:
             res = first_derivative(expr, self.dims[0], self.fd_order,
                                    side=self.side, matvec=self.transpose,
-                                   x0=self.x0)
+                                   x0=self.x0, expand=expand)
         elif len(self.dims) > 1:
             res = cross_derivative(expr, self.dims, self.fd_order, self.deriv_order,
-                                   matvec=self.transpose, x0=self.x0)
+                                   matvec=self.transpose, x0=self.x0, expand=expand)
         else:
             res = generic_derivative(expr, *self.dims, self.fd_order, self.deriv_order,
-                                     matvec=self.transpose, x0=self.x0)
+                                     matvec=self.transpose, x0=self.x0, expand=expand)
 
-        # Step 3: Evaluate remaining part of expression
-        res = res.evaluate
-
-        # Step 4: Apply substitution
+        # Step 3: Apply substitutions
         for e in self._ppsubs:
             res = res.xreplace(e)
 
