@@ -1,20 +1,21 @@
 from collections import ChainMap
+from itertools import product
 from functools import singledispatch
 
+import numpy as np
 import sympy
 from sympy.core.add import _addsort
-from sympy.core.mul import _mulsort
+from sympy.core.mul import _keep_coeff, _mulsort
 from sympy.core.decorators import call_highest_priority
 from sympy.core.evalf import evalf_table
 
 from cached_property import cached_property
 from devito.finite_differences.tools import make_shift_x0
 from devito.logger import warning
-from devito.tools import filter_ordered, flatten, split
-from devito.types.lazy import Evaluable
-from devito.types.utils import DimensionTuple
+from devito.tools import as_tuple, filter_ordered, flatten, is_integer, split
+from devito.types import Array, DimensionTuple, Evaluable, StencilDimension
 
-__all__ = ['Differentiable', 'EvalDerivative']
+__all__ = ['Differentiable', 'IndexDerivative', 'EvalDerivative']
 
 
 class Differentiable(sympy.Expr, Evaluable):
@@ -277,19 +278,36 @@ class Differentiable(sympy.Expr, Evaluable):
         from devito.finite_differences.derivative import Derivative
         return Derivative(self, *symbols, **assumptions)
 
-    def _has(self, pattern):
+    def has(self, *pattern):
         """
-        Unlike generic SymPy use cases, in Devito the majority of calls to `_has`
+        Unlike generic SymPy use cases, in Devito the majority of calls to `has`
         occur through the finite difference routines passing `sympy.core.symbol.Symbol`
         as `pattern`. Since the generic `_has` can be prohibitively expensive,
-        we here quickly handle this special case, while using the superclass' `_has`
+        we here quickly handle this special case, while using the superclass' `has`
         as fallback.
         """
-        if isinstance(pattern, type) and issubclass(pattern, sympy.Symbol):
-            # Symbols (and subclasses) are the leaves of an expression, and they
-            # are promptly available via `free_symbols`. So this is super quick
-            return any(isinstance(i, pattern) for i in self.free_symbols)
-        return super(Differentiable, self)._has(pattern)
+        for p in pattern:
+            # Following sympy convention, return True if any is found
+            if isinstance(p, type) and issubclass(p, sympy.Symbol):
+                # Symbols (and subclasses) are the leaves of an expression, and they
+                # are promptly available via `free_symbols`. So this is super quick
+                if any(isinstance(i, p) for i in self.free_symbols):
+                    return True
+        return super().has(*pattern)
+
+    def has_free(self, *patterns):
+        """
+        Return True if self has object(s) `patterns` as a free expression,
+        False otherwise.
+
+        Notes
+        -----
+        This is overridden in SymPy 1.10, but not in previous versions.
+        """
+        try:
+            return super().has_free(*patterns)
+        except AttributeError:
+            return all(i in self.free_symbols for i in patterns)
 
 
 def highest_priority(DiffOp):
@@ -360,10 +378,6 @@ class DifferentiableFunction(DifferentiableOp):
     def __new__(cls, *args, **kwargs):
         return cls.__sympy_class__.__new__(cls, *args, **kwargs)
 
-    @property
-    def evaluate(self):
-        return self.func(*[getattr(a, 'evaluate', a) for a in self.args])
-
     def _eval_at(self, func):
         return self
 
@@ -396,7 +410,7 @@ class Mul(DifferentiableOp, sympy.Mul):
         # A Mul, being a DifferentiableOp, may not trigger evaluation upon
         # construction (e.g., when an EvalDerivative is present among its
         # arguments), so here we apply a small set of basic simplifications
-        # to avoid generating functional, but also ugly, code
+        # to avoid generating functional, but ugly, code
 
         # (a*b)*c -> a*b*c (flattening)
         nested, others = split(args, lambda e: isinstance(e, Mul))
@@ -409,14 +423,30 @@ class Mul(DifferentiableOp, sympy.Mul):
         # a*1 -> a
         args = [i for i in args if i != 1]
 
+        # a*-1 -> a*-1
         # a*-1*-1 -> a
+        # a*-1*-1*-1 -> a*-1
         nminus = len([i for i in args if i == sympy.S.NegativeOne])
-        if nminus % 2 == 0:
-            args = [i for i in args if i != sympy.S.NegativeOne]
+        args = [i for i in args if i != sympy.S.NegativeOne]
+        if nminus % 2 == 1:
+            args.append(sympy.S.NegativeOne)
 
         # Reorder for homogeneity with pure SymPy types
         _mulsort(args)
 
+        # `sympy.Mul.flatten(coeff, Add)` flattens out nested Adds within Add,
+        # which would destroy `EvalDerivative`s if present. So here we perform
+        # a similar thing, but cautiously construct an evaluated Add, which
+        # will preserve the integrity of `EvalDerivative`s, if any
+        try:
+            a, b = args
+            if a.is_Rational:
+                r, b = b.as_coeff_Mul()
+                if r is sympy.S.One and type(b) is Add:
+                    return Add(*[_keep_coeff(a, bi) for bi in b.args], evaluate=False)
+        except ValueError:
+            pass
+
         return super().__new__(cls, *args, **kwargs)
 
     @property
@@ -466,10 +496,141 @@ class Mod(DifferentiableOp, sympy.Mod):
     __sympy_class__ = sympy.Mod
 
 
+class IndexSum(DifferentiableOp):
+
+    """
+    Represent the summation over a multiindex, that is a collection of
+    Dimensions, of an indexed expression.
+    """
+
+    is_commutative = True
+
+    def __new__(cls, expr, dimensions, **kwargs):
+        dimensions = as_tuple(dimensions)
+        if not dimensions:
+            return expr
+        for d in dimensions:
+            try:
+                if d.is_Dimension and is_integer(d.symbolic_size):
+                    continue
+            except AttributeError:
+                pass
+            raise ValueError("Expected Dimension with numeric size, "
+                             "got `%s` instead" % d)
+        if not expr.has_free(*dimensions):
+            raise ValueError("All Dimensions `%s` must appear in `expr` "
+                             "as free variables" % str(dimensions))
+        for i in expr.find(IndexSum):
+            for d in dimensions:
+                if d in i.dimensions:
+                    raise ValueError("Dimension `%s` already appears in a "
+                                     "nested tensor contraction" % d)
+
+        obj = sympy.Expr.__new__(cls, expr, *dimensions)
+        obj._expr = expr
+        obj._dimensions = dimensions
+
+        return obj
+
+    def __repr__(self):
+        return "%s(%s, (%s))" % (self.__class__.__name__, self.expr,
+                                 ', '.join(d.name for d in self.dimensions))
+
+    __str__ = __repr__
+
+    @property
+    def expr(self):
+        return self._expr
+
+    @property
+    def dimensions(self):
+        return self._dimensions
+
+    def _evaluate(self, **kwargs):
+        expr = self.expr._evaluate(**kwargs)
+
+        if not kwargs.get('expand', True):
+            return self.func(expr, self.dimensions)
+
+        values = product(*[list(d.range) for d in self.dimensions])
+        terms = []
+        for i in values:
+            mapper = dict(zip(self.dimensions, i))
+            terms.append(expr.xreplace(mapper))
+        return sum(terms)
+
+    @property
+    def free_symbols(self):
+        return super().free_symbols - set(self.dimensions)
+
+
+class Weights(Array):
+
+    """
+    The weights (or coefficients) of a finite-difference expansion.
+    """
+
+    def __init_finalize__(self, *args, **kwargs):
+        dimensions = as_tuple(kwargs.get('dimensions'))
+        weights = kwargs.get('initvalue')
+
+        assert len(dimensions) == 1
+        d = dimensions[0]
+        assert isinstance(d, StencilDimension) and d.symbolic_size == len(weights)
+        assert isinstance(weights, (list, tuple, np.ndarray))
+
+        kwargs['scope'] = 'static'
+
+        super().__init_finalize__(*args, **kwargs)
+
+    @property
+    def dimension(self):
+        return self.dimensions[0]
+
+    weights = Array.initvalue
+
+
+class IndexDerivative(IndexSum):
+
+    def __new__(cls, expr, dimensions, **kwargs):
+        if not (expr.is_Mul and len(expr.args) == 2):
+            raise ValueError("Expect expr*weights, got `%s` instead" % str(expr))
+        _, weights = expr.args
+        if not isinstance(weights, Weights):
+            # All of the SymPy versions we support end up placing the Weights
+            # array here, so if something changes we'll get an alarm through
+            # this exception
+            raise ValueError("Couldn't find weights array")
+
+        obj = super().__new__(cls, expr, dimensions)
+        obj._weights = weights
+
+        return obj
+
+    @property
+    def weights(self):
+        return self._weights
+
+    def _evaluate(self, **kwargs):
+        expr = super()._evaluate(**kwargs)
+
+        if not kwargs.get('expand', True):
+            return expr
+
+        w = self.weights
+        d = w.dimension
+        mapper = {w.subs(d, i): w.weights[n] for n, i in enumerate(d.range)}
+        expr = expr.xreplace(mapper)
+
+        return expr
+
+
 class EvalDerivative(DifferentiableOp, sympy.Add):
 
     is_commutative = True
 
+    __rkwargs__ = ('base',)
+
     def __new__(cls, *args, base=None, **kwargs):
         kwargs['evaluate'] = False
 
@@ -493,9 +654,7 @@ class EvalDerivative(DifferentiableOp, sympy.Add):
 
         return obj
 
-    @property
-    def func(self):
-        return lambda *a, **kw: EvalDerivative(*a, base=self.base, **kw)
+    func = DifferentiableOp._rebuild
 
     def _new_rawargs(self, *args, **kwargs):
         kwargs.pop('is_commutative', None)
@@ -584,7 +743,12 @@ def diff2sympy(expr):
             # Won't lower (e.g., EvalDerivative)
             pass
         if flag:
-            return obj.func(*args, evaluate=False), True
+            try:
+                return obj.func(*args, evaluate=False), True
+            except TypeError:
+                # In case of indices using other Function, evaluate
+                # may not be a supported argument.
+                return obj.func(*args), True
         else:
             return obj, False
 
