@@ -6,10 +6,10 @@ from cached_property import cached_property
 from devito.ir.equations import ClusterizedEq
 from devito.ir.support import (PARALLEL, PARALLEL_IF_PVT, BaseGuardBoundNext, Forward,
                                Interval, IntervalGroup, IterationSpace, DataSpace, Scope,
-                               detect_accesses, detect_io, normalize_properties)
+                               detect_accesses, detect_io, normalize_properties,
+                               normalize_syncs)
 from devito.symbolics import estimate_cost
 from devito.tools import as_tuple, flatten, frozendict
-from devito.types import normalize_syncs
 
 __all__ = ["Cluster", "ClusterGroup"]
 
@@ -23,7 +23,7 @@ class Cluster(object):
     ----------
     exprs : expr-like or list of expr-like
         An ordered sequence of expressions computing a tensor.
-    ispace : IterationSpace
+    ispace : IterationSpace, optional
         The cluster iteration space.
     guards : dict, optional
         Mapper from Dimensions to expr-like, representing the conditions under
@@ -37,7 +37,9 @@ class Cluster(object):
         Cluster asynchronously.
     """
 
-    def __init__(self, exprs, ispace, guards=None, properties=None, syncs=None):
+    def __init__(self, exprs, ispace=None, guards=None, properties=None, syncs=None):
+        ispace = ispace or IterationSpace([])
+
         self._exprs = tuple(ClusterizedEq(e, ispace=ispace) for e in as_tuple(exprs))
         self._ispace = ispace
         self._guards = frozendict(guards or {})
@@ -134,11 +136,6 @@ class Cluster(object):
     def syncs(self):
         return self._syncs
 
-    @cached_property
-    def sync_locks(self):
-        return frozendict({k: tuple(i for i in v if i.is_SyncLock)
-                           for k, v in self.syncs.items()})
-
     @cached_property
     def free_symbols(self):
         return set().union(*[e.free_symbols for e in self.exprs])
@@ -207,30 +204,43 @@ class Cluster(object):
         # Fallback to legacy is_dense checks
         return (not any(e.conditionals for e in self.exprs) and
                 not any(f.is_SparseFunction for f in self.functions) and
-                not self.is_scalar and
                 all(a.is_regular for a in self.scope.accesses))
 
+    @property
+    def is_sparse(self):
+        return not self.is_dense
+
     @cached_property
     def dtype(self):
         """
-        The arithmetic data type of the Cluster. If the Cluster performs
-        floating point arithmetic, then the expressions performing integer
-        arithmetic are ignored, assuming that they are only carrying out array
-        index calculations. If two expressions perform floating point
-        calculations with mixed precision, the data type with highest precision
-        is returned.
+        The arithmetic data type of the Cluster.
+
+        If the Cluster performs floating point arithmetic, then the expressions
+        performing integer arithmetic are ignored, assuming that they are only
+        carrying out array index calculations.
+
+        If two expressions perform calculations with different precision, the
+        data type with highest precision is returned.
         """
-        dtypes = {i.dtype for i in self.exprs}
+        dtypes = set()
+        for i in self.exprs:
+            try:
+                if np.issubdtype(i.dtype, np.generic):
+                    dtypes.add(i.dtype)
+            except TypeError:
+                # E.g. `i.dtype` is a ctypes pointer, which has no dtype equivalent
+                pass
         fdtypes = {i for i in dtypes if np.issubdtype(i, np.floating)}
+
         if len(fdtypes) > 1:
-            raise NotImplementedError("Unsupported Cluster with mixed floating "
-                                      "point arithmetic %s" % str(fdtypes))
+            return max(fdtypes, key=lambda i: np.dtype(i).itemsize)
         elif len(fdtypes) == 1:
             return fdtypes.pop()
         elif len(dtypes) == 1:
             return dtypes.pop()
         else:
-            raise ValueError("Unsupported Cluster [mixed integer arithmetic ?]")
+            # E.g., mixed integer arithmetic
+            return max(dtypes, key=lambda i: np.dtype(i).itemsize, default=None)
 
     @cached_property
     def dspace(self):
@@ -381,9 +391,9 @@ class ClusterGroup(tuple):
         return tuple(i.guards for i in self)
 
     @cached_property
-    def sync_locks(self):
-        """The synchronization locks of each Cluster in self."""
-        return tuple(i.sync_locks for i in self)
+    def syncs(self):
+        """The synchronization operations of each Cluster in self."""
+        return tuple(i.syncs for i in self)
 
     @cached_property
     def dspace(self):
@@ -393,23 +403,25 @@ class ClusterGroup(tuple):
     @cached_property
     def dtype(self):
         """
-        The arithmetic data type of this ClusterGroup. If at least one
-        Cluster performs floating point arithmetic, then Clusters performing
-        integer arithmetic are ignored. If two Clusters perform floating
-        point calculations with different precision, return the data type with
-        highest precision.
+        The arithmetic data type of this ClusterGroup.
+
+        If at least one Cluster performs floating point arithmetic, then the
+        Clusters performing integer arithmetic are ignored.
+
+        If two Clusters perform calculations with different precision, the
+        data type with highest precision is returned.
         """
         dtypes = {i.dtype for i in self}
         fdtypes = {i for i in dtypes if np.issubdtype(i, np.floating)}
         if len(fdtypes) > 1:
-            raise NotImplementedError("Unsupported ClusterGroup with mixed floating "
-                                      "point arithmetic %s" % str(fdtypes))
+            return max(fdtypes, key=lambda i: np.dtype(i).itemsize)
         elif len(fdtypes) == 1:
             return fdtypes.pop()
         elif len(dtypes) == 1:
             return dtypes.pop()
         else:
-            raise ValueError("Unsupported ClusterGroup [mixed integer arithmetic ?]")
+            # E.g., mixed integer arithmetic
+            return max(dtypes, key=lambda i: np.dtype(i).itemsize)
 
     @cached_property
     def meta(self):
