@@ -6,15 +6,16 @@ from devito.finite_differences.differentiable import diff2sympy
 from devito.ir.support import (GuardFactor, Interval, IntervalGroup, IterationSpace,
                                Stencil, detect_io, detect_accesses)
 from devito.symbolics import IntDiv, uxreplace
-from devito.tools import Pickable, frozendict
-from devito.types import Eq
+from devito.tools import Pickable, Tag, frozendict
+from devito.types import Eq, Inc, ReduceMax, ReduceMin
 
-__all__ = ['LoweredEq', 'ClusterizedEq', 'DummyEq']
+__all__ = ['LoweredEq', 'ClusterizedEq', 'DummyEq', 'OpInc', 'OpMin', 'OpMax']
 
 
-class IREq(sympy.Eq):
+class IREq(sympy.Eq, Pickable):
 
-    _state = ('is_Increment', 'ispace', 'conditionals', 'implicit_dims')
+    __rargs__ = ('lhs', 'rhs')
+    __rkwargs__ = ('ispace', 'conditionals', 'implicit_dims', 'operation')
 
     @property
     def is_Scalar(self):
@@ -22,10 +23,6 @@ class IREq(sympy.Eq):
 
     is_scalar = is_Scalar
 
-    @property
-    def is_Increment(self):
-        return self._is_Increment
-
     @property
     def ispace(self):
         return self._ispace
@@ -52,7 +49,19 @@ class IREq(sympy.Eq):
 
     @property
     def state(self):
-        return {i: getattr(self, i) for i in self._state}
+        return {i: getattr(self, i) for i in self.__rkwargs__}
+
+    @property
+    def operation(self):
+        return self._operation
+
+    @property
+    def is_Reduction(self):
+        return self.operation in (OpInc, OpMin, OpMax)
+
+    @property
+    def is_Increment(self):
+        return self.operation is OpInc
 
     def apply(self, func):
         """
@@ -64,6 +73,39 @@ class IREq(sympy.Eq):
         kwargs['conditionals'] = {k: func(v) for k, v in self.conditionals.items()}
         return self.func(*args, **kwargs)
 
+    # Pickling support
+    __reduce_ex__ = Pickable.__reduce_ex__
+
+
+class Operation(Tag):
+
+    """
+    Special operation performed by an Eq.
+    """
+
+    @classmethod
+    def detect(cls, expr):
+        reduction_mapper = {
+            Inc: OpInc,
+            ReduceMax: OpMax,
+            ReduceMin: OpMin
+        }
+        try:
+            return reduction_mapper[type(expr)]
+        except KeyError:
+            pass
+
+        # NOTE: in the future we might want to track down other kinds
+        # of operations here (e.g., memcpy). However, we don't care for
+        # now, since they would remain unexploited inside the compiler
+
+        return None
+
+
+OpInc = Operation('+')
+OpMax = Operation('max')
+OpMin = Operation('min')
+
 
 class LoweredEq(IREq):
 
@@ -74,24 +116,25 @@ class LoweredEq(IREq):
 
     A SymPy equation enriched with metadata such as an IterationSpace.
 
-    When created as ``LoweredEq(devito.Eq)``, the iteration space is automatically
-    derived from analysis of ``expr``.
+    When created as `LoweredEq(devito.Eq)`, the iteration space is automatically
+    derived from analysis of `expr`.
 
-    When created as ``LoweredEq(devito.LoweredEq, **kwargs)``, the keyword
-    arguments can be anything that appears in ``LoweredEq._state`` (e.g., ispace).
+    When created as `LoweredEq(devito.LoweredEq, **kwargs)`, the keyword
+    arguments can be anything that appears in `LoweredEq.__rkwargs__`
+    (e.g., ispace).
 
-    When created as ``LoweredEq(lhs, rhs, **kwargs)``, *all* keywords in
-    ``LoweredEq._state`` must appear in ``kwargs``.
+    When created as `LoweredEq(lhs, rhs, **kwargs)`, *all* keywords in
+    `LoweredEq.__rkwargs__` must appear in `kwargs`.
     """
 
-    _state = IREq._state + ('reads', 'writes')
+    __rkwargs__ = IREq.__rkwargs__ + ('reads', 'writes')
 
     def __new__(cls, *args, **kwargs):
         if len(args) == 1 and isinstance(args[0], LoweredEq):
             # origin: LoweredEq(devito.LoweredEq, **kwargs)
             input_expr = args[0]
             expr = sympy.Eq.__new__(cls, *input_expr.args, evaluate=False)
-            for i in cls._state:
+            for i in cls.__rkwargs__:
                 setattr(expr, '_%s' % i, kwargs.get(i) or getattr(input_expr, i))
             return expr
         elif len(args) == 1 and isinstance(args[0], Eq):
@@ -99,7 +142,7 @@ class LoweredEq(IREq):
             input_expr = expr = args[0]
         elif len(args) == 2:
             expr = sympy.Eq.__new__(cls, *args, evaluate=False)
-            for i in cls._state:
+            for i in cls.__rkwargs__:
                 setattr(expr, '_%s' % i, kwargs.pop(i))
             return expr
         else:
@@ -153,9 +196,8 @@ class LoweredEq(IREq):
         expr._ispace = ispace
         expr._conditionals = conditionals
         expr._reads, expr._writes = detect_io(expr)
-
-        expr._is_Increment = input_expr.is_Increment
         expr._implicit_dims = input_expr.implicit_dims
+        expr._operation = Operation.detect(input_expr)
 
         return expr
 
@@ -171,10 +213,10 @@ class LoweredEq(IREq):
         return LoweredEq(self.lhs.xreplace(rules), self.rhs.xreplace(rules), **self.state)
 
     def func(self, *args):
-        return super(LoweredEq, self).func(*args, **self.state, evaluate=False)
+        return self._rebuild(*args, evaluate=False)
 
 
-class ClusterizedEq(IREq, Pickable):
+class ClusterizedEq(IREq):
 
     """
     ClusterizedEq(devito.IREq, **kwargs)
@@ -185,9 +227,10 @@ class ClusterizedEq(IREq, Pickable):
     There are two main differences between a LoweredEq and a
     ClusterizedEq:
 
-    * In a ClusterizedEq, the IterationSpace must *always* be provided
-      by the caller.
-    * A ClusterizedEq is "frozen", meaning that any call to ``xreplace``
+    * To construct a ClusterizedEq, the IterationSpace must be provided
+      by the caller, whie in a LoweredEq the IterationSpace is derived
+      by analysis of the input.
+    * A ClusterizedEq is "frozen", meaning that any call to e.g. `xreplace`
       will not trigger re-evaluation (e.g., mathematical simplification)
       of the expression.
 
@@ -199,27 +242,29 @@ class ClusterizedEq(IREq, Pickable):
             # origin: ClusterizedEq(expr, **kwargs)
             input_expr = args[0]
             expr = sympy.Eq.__new__(cls, *input_expr.args, evaluate=False)
-            for i in cls._state:
-                v = kwargs[i] if i in kwargs else getattr(input_expr, i, None)
-                setattr(expr, '_%s' % i, v)
+            if isinstance(input_expr, IREq):
+                for i in cls.__rkwargs__:
+                    try:
+                        v = kwargs[i]
+                    except KeyError:
+                        v = getattr(input_expr, i, None)
+                    setattr(expr, '_%s' % i, v)
+            else:
+                expr._ispace = kwargs['ispace']
+                expr._conditionals = kwargs.get('conditionals', frozendict())
+                expr._implicit_dims = input_expr.implicit_dims
+                expr._operation = Operation.detect(input_expr)
         elif len(args) == 2:
             # origin: ClusterizedEq(lhs, rhs, **kwargs)
             expr = sympy.Eq.__new__(cls, *args, evaluate=False)
-            for i in cls._state:
+            for i in cls.__rkwargs__:
                 setattr(expr, '_%s' % i, kwargs.pop(i))
         else:
             raise ValueError("Cannot construct ClusterizedEq from args=%s "
                              "and kwargs=%s" % (str(args), str(kwargs)))
         return expr
 
-    def func(self, *args, **kwargs):
-        kwargs = {k: kwargs.get(k, v) for k, v in self.state.items()}
-        return super(ClusterizedEq, self).func(*args, **kwargs)
-
-    # Pickling support
-    _pickle_args = ['lhs', 'rhs']
-    _pickle_kwargs = IREq._state
-    __reduce_ex__ = Pickable.__reduce_ex__
+    func = IREq._rebuild
 
 
 class DummyEq(ClusterizedEq):
@@ -241,7 +286,3 @@ class DummyEq(ClusterizedEq):
         else:
             raise ValueError("Cannot construct DummyEq from args=%s" % str(args))
         return ClusterizedEq.__new__(cls, obj, ispace=obj.ispace)
-
-    # Pickling support
-    _pickle_args = ['lhs', 'rhs']
-    _pickle_kwargs = []
