@@ -2,4 +2,4 @@ from devito.ir.iet.nodes import *  # noqa
 from devito.ir.iet.visitors import *  # noqa
 from devito.ir.iet.utils import *  # noqa
 from devito.ir.iet.efunc import *  # noqa
-from devito.ir.iet.scheduler import *  # noqa
+from devito.ir.iet.algorithms import *  # noqa
