@@ -17,12 +17,14 @@ from devito.ir.iet.nodes import (Node, Iteration, Expression, ExpressionBundle,
 from devito.ir.support.space import Backward
 from devito.symbolics import ccode, uxreplace
 from devito.tools import GenericVisitor, as_tuple, filter_ordered, filter_sorted, flatten
-from devito.types.basic import AbstractFunction, Basic, IndexedData
-from devito.types import ArrayObject, Dimension, VoidPointer
+from devito.types.basic import AbstractFunction, Basic
+from devito.types import (ArrayObject, CompositeObject, Dimension, Pointer,
+                          IndexedData, DeviceMap)
 
 
 __all__ = ['FindNodes', 'FindSections', 'FindSymbols', 'MapExprStmts', 'MapNodes',
-           'IsPerfectIteration', 'printAST', 'CGen', 'Transformer', 'Uxreplace']
+           'IsPerfectIteration', 'printAST', 'CGen', 'CInterface', 'Transformer',
+           'Uxreplace']
 
 
 class Visitor(GenericVisitor):
@@ -78,10 +80,6 @@ class PrintAST(Visitor):
         body = ' %s' % str(o) if self.verbose else ''
         return self.indent + '<C.%s%s>' % (o.__class__.__name__, body)
 
-    def visit_Element(self, o):
-        body = ' %s' % str(o.element) if self.verbose else ''
-        return self.indent + '<Element%s>' % body
-
     def visit_Callable(self, o):
         self._depth += 1
         body = self._visit(o.children)
@@ -242,46 +240,74 @@ class CGen(Visitor):
 
     def visit_PointerCast(self, o):
         f = o.function
-        if isinstance(o.obj, VoidPointer):
-            obj = o.obj.name
-        elif isinstance(o.obj, ArrayObject):
-            obj = '%s->%s' % (o.obj.name, f._C_name)
-        else:
-            obj = f._C_name
+        i = f.indexed
+
         if f.is_PointerArray:
-            lvalue = c.Value(f._C_typedata, '**%s' % f.name)
-            rvalue = '(%s**) %s' % (f._C_typedata, obj)
+            # lvalue
+            lvalue = c.Value(i._C_typedata, '**%s' % f.name)
+
+            # rvalue
+            if isinstance(o.obj, ArrayObject):
+                v = '%s->%s' % (o.obj.name, f._C_name)
+            elif isinstance(o.obj, IndexedData):
+                v = f._C_name
+            else:
+                assert False
+            rvalue = '(%s**) %s' % (i._C_typedata, v)
+
         else:
+            # lvalue
+            if f.is_DiscreteFunction:
+                v = o.obj.name
+            else:
+                v = f.name
             if o.flat is None:
                 shape = ''.join("[%s]" % ccode(i) for i in o.castshape)
                 rshape = '(*)%s' % shape
-                lvalue = c.Value(f._C_typedata, '(*restrict %s)%s' % (f.name, shape))
+                lvalue = c.Value(i._C_typedata, '(*restrict %s)%s' % (v, shape))
             else:
                 rshape = '*'
-                lvalue = c.Value(f._C_typedata, '*%s' % o.flat)
+                lvalue = c.Value(i._C_typedata, '*%s' % v)
             if o.alignment:
                 lvalue = c.AlignedAttribute(f._data_alignment, lvalue)
-            if f.is_DiscreteFunction:
-                rvalue = '(%s %s) %s->%s' % (f._C_typedata, rshape, obj, f._C_field_data)
+
+            # rvalue
+            if f.is_DiscreteFunction or \
+               (f.is_Array and f._mem_mapped):
+                if isinstance(o.obj, IndexedData):
+                    v = f._C_field_data
+                elif isinstance(o.obj, DeviceMap):
+                    v = f._C_field_dmap
+                else:
+                    assert False
+
+                rvalue = '(%s %s) %s->%s' % (i._C_typedata, rshape, f._C_name, v)
             else:
-                rvalue = '(%s %s) %s' % (f._C_typedata, rshape, obj)
+                if isinstance(o.obj, Pointer):
+                    v = o.obj.name
+                else:
+                    v = f._C_name
+
+                rvalue = '(%s %s) %s' % (i._C_typedata, rshape, v)
+
         return c.Initializer(lvalue, rvalue)
 
     def visit_Dereference(self, o):
         a0, a1 = o.functions
         if a1.is_PointerArray or a1.is_TempFunction:
+            i = a1.indexed
             if o.flat is None:
                 shape = ''.join("[%s]" % ccode(i) for i in a0.symbolic_shape[1:])
-                rvalue = '(%s (*)%s) %s[%s]' % (a1._C_typedata, shape, a1.name,
+                rvalue = '(%s (*)%s) %s[%s]' % (i._C_typedata, shape, a1.name,
                                                 a1.dim.name)
                 lvalue = c.AlignedAttribute(
                     a0._data_alignment,
-                    c.Value(a0._C_typedata, '(*restrict %s)%s' % (a0.name, shape))
+                    c.Value(i._C_typedata, '(*restrict %s)%s' % (a0.name, shape))
                 )
             else:
-                rvalue = '(%s *) %s[%s]' % (a1._C_typedata, a1.name, a1.dim.name)
+                rvalue = '(%s *) %s[%s]' % (i._C_typedata, a1.name, a1.dim.name)
                 lvalue = c.AlignedAttribute(
-                    a0._data_alignment, c.Value(a0._C_typedata, '*restrict %s' % o.flat)
+                    a0._data_alignment, c.Value(i._C_typedata, '*restrict %s' % a0.name)
                 )
         else:
             rvalue = '%s->%s' % (a1.name, a0._C_name)
@@ -306,32 +332,37 @@ class CGen(Visitor):
             footer = [c.Comment("End %s" % o.name)]
         return c.Module(header + body + footer)
 
-    def visit_Element(self, o):
-        return o.element
+    def visit_Return(self, o):
+        v = 'return'
+        if o.value is not None:
+            v += ' %s' % o.value
+        return c.Statement(v)
 
     def visit_Definition(self, o):
         f = o.function
 
-        if f.is_AbstractFunction:
-            if f.is_PointerArray:
-                v = "*"
-            else:
-                v = ""
-            if o.shape is None:
-                v = "%s*%s" % (v, f._C_name)
-            else:
-                v = "%s%s%s" % (v, f._C_name, o.shape)
-            if o.qualifier is not None:
-                v = "%s %s" % (v, o.qualifier)
-        elif f.is_LocalObject:
-            v = f._C_name
+        if o.shape is not None:
+            v = o.shape
         else:
-            assert False
+            v = ""
 
-        if o.constructor_args:
-            v = MultilineCall(v, o.constructor_args, True)
+        if o.qualifier is not None:
+            v = "%s %s" % (v, o.qualifier)
 
-        v = c.Value(f._C_typedata, v)
+        v = "%s%s" % (f._C_name, v)
+
+        if f._mem_stack:
+            v = c.Value(f._C_typedata, v)
+        else:
+            if o.cargs:
+                v = MultilineCall(v, o.cargs, True)
+
+            # Aesthetics: `float * a` -> `float *a`
+            try:
+                t, stars = f._C_typename.split()
+                v = c.Value(t, '%s%s' % (stars, v))
+            except ValueError:
+                v = c.Value(f._C_typename, v)
 
         if o.initvalue is not None:
             v = c.Initializer(v, o.initvalue)
@@ -367,10 +398,7 @@ class CGen(Visitor):
             return MultilineCall(o.name, arguments, nested_call, o.is_indirect, cast)
         else:
             call = MultilineCall(o.name, arguments, True, o.is_indirect, cast)
-            if retobj.is_AbstractFunction:
-                return c.Initializer(c.Value(retobj._C_typename, retobj._C_name), call)
-            else:
-                return c.Initializer(c.Value(retobj._C_typedata, ccode(retobj)), call)
+            return c.Initializer(c.Value(retobj._C_typename, retobj._C_name), call)
 
     def visit_Conditional(self, o):
         try:
@@ -467,9 +495,35 @@ class CGen(Visitor):
         body = flatten(self._visit(i) for i in o.children)
         return c.Collection(body)
 
-    def visit_Operator(self, o):
-        blankline = c.Line("")
+    # Operator-handle machinery
 
+    def _operator_includes(self, o):
+        return [c.Include(i, system=(False if i.endswith('.h') else True))
+                for i in o._includes]
+
+    def _operator_typedecls(self, o, mode='all'):
+        if mode == 'all':
+            xfilter = lambda i: True
+        else:
+            public_types = (AbstractFunction, CompositeObject)
+            if mode == 'public':
+                xfilter = lambda i: isinstance(i, public_types)
+            else:
+                xfilter = lambda i: not isinstance(i, public_types)
+
+        candidates = o.parameters + tuple(o._dspace.parts)
+        typedecls = [i._C_typedecl for i in candidates
+                     if xfilter(i) and i._C_typedecl is not None]
+        for i in o._func_table.values():
+            if not i.local:
+                continue
+            typedecls.extend([j._C_typedecl for j in i.root.parameters
+                              if xfilter(j) and j._C_typedecl is not None])
+        typedecls = filter_sorted(typedecls, key=lambda i: i.tpname)
+
+        return typedecls
+
+    def visit_Operator(self, o, mode='all'):
         # Kernel signature and body
         body = flatten(self._visit(i) for i in o.children)
         decls = self._args_decl(o.parameters)
@@ -487,25 +541,50 @@ class CGen(Visitor):
                                                     self._args_decl(i.root.parameters)))
                 efuncs.extend([self._visit(i.root), blankline])
 
-        # Header files, extra definitions, ...
-        header = [c.Define(*i) for i in o._headers] + [blankline]
-        includes = [c.Include(i, system=(False if i.endswith('.h') else True))
-                    for i in o._includes]
-        includes += [blankline]
-        cdefs = [i._C_typedecl for i in o.parameters if i._C_typedecl is not None]
-        for i in o._func_table.values():
-            if i.local:
-                cdefs.extend([j._C_typedecl for j in i.root.parameters
-                              if j._C_typedecl is not None])
-        cdefs = filter_sorted(cdefs, key=lambda i: i.tpname)
-        if o._compiler.src_ext in ('cpp', 'cu'):
-            cdefs += [c.Extern('C', signature)]
-        cdefs = [i for j in cdefs for i in (j, blankline)]
+        # Definitions
+        headers = [c.Define(*i) for i in o._headers] + [blankline]
 
-        return c.Module(header + includes + cdefs +
+        # Header files
+        includes = self._operator_includes(o) + [blankline]
+
+        # Type declarations
+        typedecls = self._operator_typedecls(o, mode)
+        if mode in ('all', 'public') and o._compiler.src_ext in ('cpp', 'cu'):
+            typedecls.append(c.Extern('C', signature))
+        typedecls = [i for j in typedecls for i in (j, blankline)]
+
+        return c.Module(headers + includes + typedecls +
                         esigns + [blankline, kernel] + efuncs)
 
 
+class CInterface(CGen):
+
+    def _operator_includes(self, o):
+        includes = super()._operator_includes(o)
+        includes.append(c.Include("%s.h" % o.name, system=False))
+
+        return includes
+
+    def visit_Operator(self, o):
+        # Generate the code for the cfile
+        ccode = super().visit_Operator(o, mode='private')
+
+        # Generate the code for the hfile
+        typedecls = self._operator_typedecls(o, mode='public')
+        guarded_typedecls = []
+        for i in typedecls:
+            guard = "DEVITO_%s" % i.tpname.upper()
+            iflines = [c.Define(guard, ""), blankline, i, blankline]
+            guarded_typedecl = c.IfNDef(guard, iflines, [])
+            guarded_typedecls.extend([guarded_typedecl, blankline])
+
+        decls = self._args_decl(o.parameters)
+        signature = c.FunctionDeclaration(c.Value(o.retval, o.name), decls)
+        hcode = c.Module(guarded_typedecls + [blankline, signature, blankline])
+
+        return ccode, hcode
+
+
 class FindSections(Visitor):
 
     @classmethod
@@ -890,6 +969,23 @@ class Uxreplace(Transformer):
     def visit_Expression(self, o):
         return o._rebuild(expr=uxreplace(o.expr, self.mapper))
 
+    def visit_Definition(self, o):
+        try:
+            return o._rebuild(function=self.mapper[o.function])
+        except KeyError:
+            return o
+
+    def visit_Return(self, o):
+        try:
+            return o._rebuild(value=self.mapper[o.value])
+        except KeyError:
+            return o
+
+    def visit_Callable(self, o):
+        body = self._visit(o.body)
+        parameters = [self.mapper.get(i, i) for i in o.parameters]
+        return o._rebuild(body=body, parameters=parameters)
+
     def visit_Call(self, o):
         arguments = [uxreplace(i, self.mapper) for i in o.arguments]
         return o._rebuild(arguments=arguments)
@@ -900,11 +996,26 @@ class Uxreplace(Transformer):
         else_body = self._visit(o.else_body)
         return o._rebuild(condition=condition, then_body=then_body, else_body=else_body)
 
+    def visit_PragmaTransfer(self, o):
+        function = uxreplace(o.function, self.mapper)
+        arguments = [uxreplace(i, self.mapper) for i in o.arguments]
+        return o._rebuild(function=function, arguments=arguments)
+
+    def visit_HaloSpot(self, o):
+        hs = o.halo_scheme
+        fmapper = {self.mapper.get(k, k): v for k, v in hs.fmapper.items()}
+        halo_scheme = hs.build(fmapper, hs.honored)
+        body = self._visit(o.body)
+        return o._rebuild(halo_scheme=halo_scheme, body=body)
+
     visit_ThreadedProdder = visit_Call
 
 
 # Utils
 
+blankline = c.Line("")
+
+
 def printAST(node, verbose=True):
     return PrintAST(verbose=verbose)._visit(node)
 
