@@ -153,6 +153,14 @@ class Interval(AbstractInterval):
     def _rebuild(self):
         return Interval(self.dim, self.lower, self.upper, self.stamp)
 
+    @property
+    def symbolic_min(self):
+        return self.dim.symbolic_min + self.lower
+
+    @property
+    def symbolic_max(self):
+        return self.dim.symbolic_max + self.upper
+
     @cached_property
     def size(self):
         """
@@ -192,8 +200,8 @@ class Interval(AbstractInterval):
             npoints /= n
         else:
             # Typically we end up here (Dimension, SubDimension, BlockDimension)
-            upper_extreme = self.dim.symbolic_max + self.upper
-            lower_extreme = self.dim.symbolic_min + self.lower
+            upper_extreme = self.symbolic_max
+            lower_extreme = self.symbolic_min
 
             npoints = (upper_extreme - lower_extreme + 1)
 
@@ -311,7 +319,7 @@ class IntervalGroup(PartialOrderTuple):
 
     @cached_property
     def dimensions(self):
-        return filter_ordered([i.dim for i in self])
+        return tuple(filter_ordered([i.dim for i in self]))
 
     @property
     def size(self):
@@ -540,6 +548,8 @@ class IterationInterval(object):
         return "%s%s" % (self.interval, self.direction)
 
     def __eq__(self, other):
+        if not isinstance(other, IterationInterval):
+            return False
         return self.direction is other.direction and self.interval == other.interval
 
     def __hash__(self):
@@ -602,7 +612,7 @@ class Space(object):
 
     @cached_property
     def dimensions(self):
-        return filter_ordered(self.intervals.dimensions)
+        return tuple(filter_ordered(self.intervals.dimensions))
 
     @property
     def size(self):
@@ -730,7 +740,11 @@ class IterationSpace(Space):
                      self.directions))
 
     def __getitem__(self, key):
-        return self.intervals[key]
+        retval = self.intervals[key]
+        if isinstance(key, slice):
+            return self.project(lambda d: d in retval.dimensions)
+        else:
+            return retval
 
     @classmethod
     def union(cls, *others):
@@ -760,6 +774,9 @@ class IterationSpace(Space):
 
         return IterationSpace(intervals, sub_iterators, directions)
 
+    def index(self, key):
+        return self.intervals.index(key)
+
     def add(self, other):
         return IterationSpace(self.intervals.add(other), self.sub_iterators,
                               self.directions)
@@ -845,9 +862,9 @@ class IterationSpace(Space):
 
     def promote(self, cond):
         intervals = self.intervals.promote(cond)
-        sub_iterators = {i.promote(cond).dim: self.sub_iterators.get(i.dim, ())
+        sub_iterators = {i.promote(cond).dim: self.sub_iterators[i.dim]
                          for i in self.intervals}
-        directions = {i.promote(cond).dim: self.directions.get(i.dim, ())
+        directions = {i.promote(cond).dim: self.directions[i.dim]
                       for i in self.intervals}
 
         return IterationSpace(intervals, sub_iterators, directions)
@@ -891,7 +908,7 @@ class IterationSpace(Space):
     @cached_property
     def dimensions(self):
         sub_dims = flatten(i._defines for v in self.sub_iterators.values() for i in v)
-        return filter_ordered(self.itdimensions + sub_dims)
+        return tuple(filter_ordered(self.itdimensions + tuple(sub_dims)))
 
     @cached_property
     def nonderived_directions(self):
