@@ -373,7 +373,7 @@ def classify(exprs, ispace):
                         v[(d, LEFT)] = STENCIL
                         v[(d, RIGHT)] = STENCIL
                 else:
-                    v[(d, i.aindices[d])] = NONE
+                    v[(d, i[d])] = NONE
 
             # Does `i` actually require a halo exchange?
             if not any(hl is STENCIL for hl in v.values()):
@@ -425,8 +425,21 @@ def classify(exprs, ispace):
                 # Max or Min is the same since `d` isn't an `ispace` Dimension
                 func = Max
             candidates = [i for i in aindices if not is_integer(i)]
-            candidates = {(i.origin if d.is_Stepping else i) - d: i for i in candidates}
-            loc_indices[d] = candidates[func(*candidates.keys())]
+            m = {}
+            for i in candidates:
+                try:
+                    k = i.origin - d
+                except AttributeError:
+                    # E.g., `i=otime`, that is a plain Symbol, or `i` not a
+                    # SteppingDimension
+                    k = i - d
+                m[k] = i
+            try:
+                loc_indices[d] = m[func(*m.keys())]
+            except KeyError:
+                # E.g., `aindices = [0, 1, d+1]` -- it doesn't really matter
+                # what we put here, so we place 0 as it's the old behaviour
+                loc_indices[d] = 0
 
         mapper[f] = HaloSchemeEntry(frozendict(loc_indices), frozenset(halos))
 
