@@ -1,4 +1,4 @@
-from collections import OrderedDict
+from collections import OrderedDict, namedtuple
 from operator import attrgetter
 from math import ceil
 
@@ -9,18 +9,18 @@ import numpy as np
 from devito.arch import compiler_registry, platform_registry
 from devito.data import default_allocator
 from devito.exceptions import InvalidOperator
-from devito.logger import info, perf, warning, is_log_enabled_for
-from devito.ir.equations import LoweredEq, lower_exprs, generate_implicit_exprs
+from devito.logger import debug, info, perf, warning, is_log_enabled_for
+from devito.ir.equations import LoweredEq, lower_exprs
 from devito.ir.clusters import ClusterGroup, clusterize
-from devito.ir.iet import (Callable, EntryFunction, FindSymbols, MetaCall,
+from devito.ir.iet import (Callable, CInterface, EntryFunction, FindSymbols, MetaCall,
                            derive_parameters, iet_build)
+from devito.ir.support import SymbolRegistry
 from devito.ir.stree import stree_build
 from devito.operator.profiling import create_profile
 from devito.operator.registry import operator_selector
-from devito.operator.symbols import SymbolRegistry
 from devito.mpi import MPI
 from devito.parameters import configuration
-from devito.passes import Graph, instrument
+from devito.passes import Graph, generate_implicit, instrument
 from devito.symbolics import estimate_cost
 from devito.tools import (DAG, Signer, ReducerMap, as_tuple, flatten, filter_sorted,
                           split, timed_pass, timed_region)
@@ -151,9 +151,6 @@ class Operator(Callable):
         # Normalize input arguments for the selected Operator
         kwargs = cls._normalize_kwargs(**kwargs)
 
-        # Create a symbol registry
-        kwargs['sregistry'] = SymbolRegistry()
-
         # Lower to a JIT-compilable object
         with timed_region('op-compile') as r:
             op = cls._build(expressions, **kwargs)
@@ -170,29 +167,14 @@ class Operator(Callable):
 
     @classmethod
     def _build(cls, expressions, **kwargs):
-        expressions = as_tuple(expressions)
-
-        # Input check
-        if any(not isinstance(i, Evaluable) for i in expressions):
-            raise InvalidOperator("Only `devito.Evaluable` are allowed.")
-
-        # Python-level (i.e., compile time) and C-level (i.e., run time) performance
+        # Python- (i.e., compile-) and C-level (i.e., run-time) performance
         profiler = create_profile('timers')
 
-        # Lower input expressions
-        expressions = cls._lower_exprs(expressions, **kwargs)
-
-        # Group expressions based on iteration spaces and data dependences
-        clusters = cls._lower_clusters(expressions, profiler, **kwargs)
-
-        # Lower Clusters to a ScheduleTree
-        stree = cls._lower_stree(clusters, **kwargs)
-
-        # Lower ScheduleTree to an Iteration/Expression Tree
-        iet, byproduct = cls._lower_iet(stree, profiler, **kwargs)
+        # Lower the input expressions into an IET
+        irs, byproduct = cls._lower(expressions, profiler=profiler, **kwargs)
 
         # Make it an actual Operator
-        op = Callable.__new__(cls, **iet.args)
+        op = Callable.__new__(cls, **irs.iet.args)
         Callable.__init__(op, **op.args)
 
         # Header files, etc.
@@ -209,6 +191,8 @@ class Operator(Callable):
         op._cfunction = None
 
         # Potentially required for lazily allocated Functions
+        op._mode = kwargs['mode']
+        op._options = kwargs['options']
         op._allocator = kwargs['allocator']
         op._platform = kwargs['platform']
 
@@ -218,15 +202,15 @@ class Operator(Callable):
                                            for i in profiler._ext_calls]))
         op._func_table.update(OrderedDict([(i.root.name, i) for i in byproduct.funcs]))
 
-        # Internal state. May be used to store information about previous runs,
-        # autotuning reports, etc
+        # Internal mutable state to store information about previous runs, autotuning
+        # reports, etc
         op._state = cls._initialize_state(**kwargs)
 
         # Produced by the various compilation passes
-        op._input = filter_sorted(flatten(e.reads + e.writes for e in expressions))
-        op._output = filter_sorted(flatten(e.writes for e in expressions))
-        op._dimensions = set().union(*[e.dimensions for e in expressions])
-        op._dtype, op._dspace = clusters.meta
+        op._input = filter_sorted(flatten(e.reads + e.writes for e in irs.expressions))
+        op._output = filter_sorted(flatten(e.writes for e in irs.expressions))
+        op._dimensions = set().union(*[e.dimensions for e in irs.expressions])
+        op._dtype, op._dspace = irs.clusters.meta
         op._profiler = profiler
 
         return op
@@ -237,9 +221,45 @@ class Operator(Callable):
 
     # Compilation -- Expression level
 
+    @classmethod
+    def _lower(cls, expressions, **kwargs):
+        """
+        Perform the lowering Expressions -> Clusters -> ScheduleTree -> IET.
+        """
+        # Create a symbol registry
+        kwargs['sregistry'] = SymbolRegistry()
+
+        expressions = as_tuple(expressions)
+
+        # Input check
+        if any(not isinstance(i, Evaluable) for i in expressions):
+            raise InvalidOperator("Only `devito.Evaluable` are allowed.")
+
+        # Enable recursive lowering
+        # This may be used by a compilation pass that constructs a new
+        # expression for which a partial or complete lowering is desired
+        kwargs['lower'] = cls._lower
+
+        # [Eq] -> [LoweredEq]
+        expressions = cls._lower_exprs(expressions, **kwargs)
+
+        # [LoweredEq] -> [Clusters]
+        clusters = cls._lower_clusters(expressions, **kwargs)
+
+        # [Clusters] -> ScheduleTree
+        stree = cls._lower_stree(clusters, **kwargs)
+
+        # ScheduleTree -> unbounded IET
+        uiet = cls._lower_uiet(stree, **kwargs)
+
+        # unbounded IET -> IET
+        iet, byproduct = cls._lower_iet(uiet, **kwargs)
+
+        return IRs(expressions, clusters, stree, uiet, iet), byproduct
+
     @classmethod
     def _initialize_state(cls, **kwargs):
-        return {'optimizations': kwargs.get('mode', configuration['opt'])}
+        return {}
 
     @classmethod
     def _specialize_dsl(cls, expressions, **kwargs):
@@ -265,7 +285,6 @@ class Operator(Callable):
         """
         Expression lowering:
 
-            * Form and gather any required implicit expressions;
             * Apply rewrite rules;
             * Evaluate derivatives;
             * Flatten vectorial equations;
@@ -273,9 +292,6 @@ class Operator(Callable):
             * Apply substitution rules;
             * Shift indices for domain alignment.
         """
-        # Add in implicit expressions
-        expressions = generate_implicit_exprs(expressions)
-
         # Specialization is performed on unevaluated expressions
         expressions = cls._specialize_dsl(expressions, **kwargs)
 
@@ -304,7 +320,7 @@ class Operator(Callable):
 
     @classmethod
     @timed_pass(name='lowering.Clusters')
-    def _lower_clusters(cls, expressions, profiler, **kwargs):
+    def _lower_clusters(cls, expressions, profiler=None, **kwargs):
         """
         Clusters lowering:
 
@@ -312,7 +328,10 @@ class Operator(Callable):
             * Introduce guards for conditional Clusters;
             * Analyze Clusters to detect computational properties such
               as parallelism.
+            * Optimize Clusters for performance
         """
+        sregistry = kwargs['sregistry']
+
         # Build a sequence of Clusters from a sequence of Eqs
         clusters = clusterize(expressions, **kwargs)
 
@@ -323,7 +342,13 @@ class Operator(Callable):
 
         # Operation count after specialization
         final_ops = sum(estimate_cost(c.exprs) for c in clusters if c.is_dense)
-        profiler.record_ops_variation(init_ops, final_ops)
+        try:
+            profiler.record_ops_variation(init_ops, final_ops)
+        except AttributeError:
+            pass
+
+        # Generate implicit Clusters from higher level abstractions
+        clusters = generate_implicit(clusters, sregistry=sregistry)
 
         return ClusterGroup(clusters)
 
@@ -347,7 +372,7 @@ class Operator(Callable):
             * Derive sections for performance profiling
         """
         # Build a ScheduleTree from a sequence of Clusters
-        stree = stree_build(clusters)
+        stree = stree_build(clusters, **kwargs)
 
         stree = cls._specialize_stree(stree)
 
@@ -362,13 +387,31 @@ class Operator(Callable):
         """
         return graph
 
+    @classmethod
+    @timed_pass(name='lowering.uIET')
+    def _lower_uiet(cls, stree, profiler=None, **kwargs):
+        """
+        Turn a ScheduleTree into an unbounded Iteration/Expression tree, that is
+        in essence a "floating" IET where one or more variables may be unbounded
+        (i.e., no definition placed yet).
+        """
+        # Build an unbounded IET from a ScheduleTree
+        uiet = iet_build(stree)
+
+        # Analyze the IET Sections for C-level profiling
+        try:
+            profiler.analyze(uiet)
+        except AttributeError:
+            pass
+
+        return uiet
+
     @classmethod
     @timed_pass(name='lowering.IET')
-    def _lower_iet(cls, stree, profiler, **kwargs):
+    def _lower_iet(cls, uiet, profiler=None, **kwargs):
         """
         Iteration/Expression tree lowering:
 
-            * Turn a ScheduleTree into an Iteration/Expression tree;
             * Introduce distributed-memory, shared-memory, and SIMD parallelism;
             * Introduce optimizations for data locality;
             * Finalize (e.g., symbol definitions, array casts)
@@ -376,16 +419,10 @@ class Operator(Callable):
         name = kwargs.get("name", "Kernel")
         sregistry = kwargs['sregistry']
 
-        # Build an IET from a ScheduleTree
-        iet = iet_build(stree)
-
-        # Analyze the IET Sections for C-level profiling
-        profiler.analyze(iet)
-
         # Wrap the IET with an EntryFunction (a special Callable representing
         # the entry point of the generated library)
-        parameters = derive_parameters(iet, True)
-        iet = EntryFunction(name, iet, 'int', parameters, ())
+        parameters = derive_parameters(uiet, True)
+        iet = EntryFunction(name, uiet, 'int', parameters, ())
 
         # Lower IET to a target-specific IET
         graph = Graph(iet)
@@ -489,7 +526,7 @@ class Operator(Callable):
 
         # An ArgumentsMap carries additional metadata that may be used by
         # the subsequent phases of the arguments processing
-        args = kwargs['args'] = ArgumentsMap(args, grid, self._allocator, self._platform)
+        args = kwargs['args'] = ArgumentsMap(args, grid, self)
 
         # Process Dimensions
         # A topological sorting is used so that derived Dimensions are processed after
@@ -504,6 +541,12 @@ class Operator(Callable):
         for o in self.objects:
             args.update(o._arg_values(grid=grid, **kwargs))
 
+        # In some "lower-level" Operators implementing a random piece of C, such as
+        # one or more calls to third-party library functions, there could still be
+        # at this point unprocessed arguments (e.g., scalars)
+        kwargs.pop('args')
+        args.update({k: v for k, v in kwargs.items() if k not in args})
+
         # Sanity check
         for p in self.parameters:
             p._arg_check(args, self._dspace[p])
@@ -546,6 +589,7 @@ class Operator(Callable):
                 pass
         for d in self.dimensions:
             ret.update(d._arg_names)
+        ret.update(p.name for p in self.parameters)
         return frozenset(ret)
 
     def _autotune(self, args, setup):
@@ -561,7 +605,7 @@ class Operator(Callable):
                 raise ValueError("No value found for parameter %s" % p.name)
         return args
 
-    # JIT compilation
+    # Code generation and JIT compilation
 
     @cached_property
     def _soname(self):
@@ -603,6 +647,39 @@ class Operator(Callable):
 
         return self._cfunction
 
+    def cinterface(self, force=False):
+        """
+        Generate two files under the prescribed temporary directory:
+
+            * `X.c` (or `X.cpp`): the code generated for this Operator;
+            * `X.h`: an header file representing the interface of `X.c`.
+
+        Where `X=self.name`.
+
+        Parameters
+        ----------
+        force : bool, optional
+            Overwrite any existing files. Defaults to False.
+        """
+        dest = self._compiler.get_jit_dir()
+        name = dest.joinpath(self.name)
+
+        cfile = name.with_suffix(".%s" % self._compiler.src_ext)
+        hfile = name.with_suffix('.h')
+
+        # Generate the .c and .h code
+        ccode, hcode = CInterface().visit(self)
+
+        for f, code in [(cfile, ccode), (hfile, hcode)]:
+            if not force and f.is_file():
+                debug("`%s` was not saved in `%s` as it already exists" % (f.name, dest))
+            else:
+                with open(str(f), 'w') as ff:
+                    ff.write(str(code))
+                debug("`%s` successfully saved in `%s`" % (f.name, dest))
+
+        return ccode, hcode
+
     # Execution
 
     def __call__(self, **kwargs):
@@ -801,8 +878,7 @@ class Operator(Callable):
                     if a in args:
                         perf_args[a] = args[a]
                         break
-        perf("Performance[mode=%s] arguments: %s" % (self._state['optimizations'],
-                                                     perf_args))
+        perf("Performance[mode=%s] arguments: %s" % (self._mode, perf_args))
 
         return summary
 
@@ -848,14 +924,19 @@ class Operator(Callable):
 # Misc helpers
 
 
+IRs = namedtuple('IRs', 'expressions clusters stree uiet iet')
+
+
 class ArgumentsMap(dict):
 
-    def __init__(self, args, grid, allocator, platform):
+    def __init__(self, args, grid, op):
         super().__init__(args)
 
         self.grid = grid
-        self.allocator = allocator
-        self.platform = platform
+
+        self.allocator = op._allocator
+        self.platform = op._platform
+        self.options = op._options
 
         # Compute total used memory
         self.memused = sum(v.nbytes for v in args.values() if isinstance(v, np.ndarray))
@@ -968,11 +1049,13 @@ def parse_kwargs(**kwargs):
         if compiler not in configuration._accepted['compiler']:
             raise InvalidOperator("Illegal `compiler=%s`" % str(compiler))
         kwargs['compiler'] = compiler_registry[compiler](platform=kwargs['platform'],
-                                                         language=kwargs['language'])
+                                                         language=kwargs['language'],
+                                                         mpi=configuration['mpi'])
     elif any([platform, language]):
         kwargs['compiler'] =\
             configuration['compiler'].__new_with__(platform=kwargs['platform'],
-                                                   language=kwargs['language'])
+                                                   language=kwargs['language'],
+                                                   mpi=configuration['mpi'])
     else:
         kwargs['compiler'] = configuration['compiler'].__new_with__()
 
