@@ -1,9 +1,9 @@
-from devito.symbolics.queries import (q_indexed, q_function, q_terminal, q_leaf, q_xop,
-                                      q_scalar)
+from devito.symbolics.queries import (q_indexed, q_function, q_terminal, q_leaf,
+                                      q_symbol, q_dimension)
 from devito.tools import as_tuple
 
 __all__ = ['retrieve_indexed', 'retrieve_functions', 'retrieve_function_carriers',
-           'retrieve_terminals', 'retrieve_xops', 'retrieve_scalars', 'search']
+           'retrieve_terminals', 'retrieve_symbols', 'retrieve_dimensions', 'search']
 
 
 class Search(object):
@@ -70,22 +70,6 @@ class Search(object):
             found.update(self.collection.wrap(expr))
         return found
 
-    def dfs_first_hit(self, expr):
-        """
-        Perform a DFS search, returning immediately when a node matches the query.
-
-        Parameters
-        ----------
-        expr : expr-like
-            The searched expression.
-        """
-        found = self.collection()
-        for a in self._next(expr):
-            found.update(self.dfs_first_hit(a))
-        if not found and self.query(expr):
-            found.update(self.collection.wrap(expr))
-        return found
-
     def bfs(self, expr):
         """
         Perform a BFS search.
@@ -131,8 +115,6 @@ def search(exprs, query, mode='unique', visit='dfs', deep=False):
     for e in as_tuple(exprs):
         if visit == 'dfs':
             found.update(searcher.dfs(e))
-        elif visit == 'dfs_first_hit':
-            found.update(searcher.dfs_first_hit(e))
         elif visit == 'bfs':
             found.update(searcher.bfs(e))
         elif visit == "bfs_first_hit":
@@ -156,9 +138,9 @@ def retrieve_functions(exprs, mode='all'):
     return search(exprs, q_function, mode, 'dfs')
 
 
-def retrieve_scalars(exprs, mode='all'):
+def retrieve_symbols(exprs, mode='all'):
     """Shorthand to retrieve the Scalar in ``exprs``."""
-    return search(exprs, q_scalar, mode, 'dfs')
+    return search(exprs, q_symbol, mode, 'dfs')
 
 
 def retrieve_function_carriers(exprs, mode='all'):
@@ -185,6 +167,6 @@ def retrieve_terminals(exprs, mode='all', deep=False):
     return search(exprs, q_terminal, mode, 'dfs', deep)
 
 
-def retrieve_xops(exprs):
-    """Shorthand to retrieve the arithmetic operations within ``exprs``."""
-    return search(exprs, q_xop, 'all', 'dfs')
+def retrieve_dimensions(exprs, mode='all', deep=False):
+    """Shorthand to retrieve the dimensions in ``exprs``."""
+    return search(exprs, q_dimension, mode, 'dfs', deep)
