@@ -1,15 +1,15 @@
-from ctypes import POINTER, Structure
+from ctypes import POINTER, Structure, c_void_p, c_ulong
 from math import ceil
 
 import numpy as np
 from cached_property import cached_property
-from cgen import Struct, Value
 
 from devito.parameters import configuration
-from devito.tools import as_tuple, ctypes_to_cstr, dtype_to_ctype
+from devito.tools import as_tuple, c_restrict_void_p, dtype_to_ctype
 from devito.types.basic import AbstractFunction, IndexedData
+from devito.types.utils import CtypesFactory
 
-__all__ = ['Array', 'ArrayObject', 'PointerArray']
+__all__ = ['Array', 'ArrayMapped', 'ArrayObject', 'PointerArray', 'IndexedArray']
 
 
 class ArrayBasic(AbstractFunction):
@@ -29,10 +29,6 @@ class ArrayBasic(AbstractFunction):
         else:
             return super()._C_name
 
-    @property
-    def _C_ctype(self):
-        return POINTER(dtype_to_ctype(self.dtype))
-
     @property
     def _C_aliases(self):
         return (self, self.indexed)
@@ -77,12 +73,17 @@ class Array(ArrayBasic):
         The halo region of the object.
     padding : iterable of 2-tuples, optional
         The padding region of the object.
+    liveness : str, optional
+        The liveness of the object. Allowed values: 'eager', 'lazy'. Defaults
+        to 'lazy'. Used to override `_mem_internal_eager` and `_mem_internal_lazy`.
     space : str, optional
-        The memory space. Allowed values: 'local', 'mapped'.  Defaults to 'local'.
-        Used to override `_mem_local` and `_mem_mapped`.
+        The memory space. Allowed values: 'local', 'mapped', 'host'. Defaults
+        to 'local'. Used to override `_mem_local` and `_mem_mapped`.
     scope : str, optional
-        The scope in the given memory space. Allowed values: 'heap', 'stack'.
-        Defaults to 'heap'. This may not have an impact on certain platforms.
+        The scope in the given memory space. Allowed values: 'heap', 'stack',
+        'static'. Defaults to 'heap'. 'static' means a static array in a
+        C/C++ sense and, therefore, implies 'stack'.
+        Note: not all scopes make sense for a given space.
     initvalue : array-like, optional
         The initial content of the Array. Must be None if `scope='heap'`.
 
@@ -94,18 +95,29 @@ class Array(ArrayBasic):
 
     is_Array = True
 
+    __rkwargs__ = (AbstractFunction.__rkwargs__ +
+                   ('dimensions', 'liveness', 'space', 'scope', 'initvalue'))
+
     def __new__(cls, *args, **kwargs):
         kwargs.update({'options': {'evaluate': False}})
-        return AbstractFunction.__new__(cls, *args, **kwargs)
+        space = kwargs.get('space', 'local')
+
+        if cls is Array and space == 'mapped':
+            return AbstractFunction.__new__(ArrayMapped, *args, **kwargs)
+        else:
+            return AbstractFunction.__new__(cls, *args, **kwargs)
 
     def __init_finalize__(self, *args, **kwargs):
         super(Array, self).__init_finalize__(*args, **kwargs)
 
+        self._liveness = kwargs.get('liveness', 'lazy')
+        assert self._liveness in ['eager', 'lazy']
+
         self._space = kwargs.get('space', 'local')
-        assert self._space in ['local', 'remote', 'mapped']
+        assert self._space in ['local', 'mapped', 'host']
 
         self._scope = kwargs.get('scope', 'heap')
-        assert self._scope in ['heap', 'stack']
+        assert self._scope in ['heap', 'stack', 'static']
 
         self._initvalue = kwargs.get('initvalue')
         assert self._initvalue is None or self._scope != 'heap'
@@ -148,8 +160,8 @@ class Array(ArrayBasic):
         return kwargs.get('dtype', np.float32)
 
     @property
-    def _C_typename(self):
-        return ctypes_to_cstr(self._C_ctype)
+    def liveness(self):
+        return self._liveness
 
     @property
     def space(self):
@@ -159,6 +171,18 @@ class Array(ArrayBasic):
     def scope(self):
         return self._scope
 
+    @property
+    def _C_ctype(self):
+        return POINTER(dtype_to_ctype(self.dtype))
+
+    @property
+    def _mem_internal_eager(self):
+        return self._liveness == 'eager'
+
+    @property
+    def _mem_internal_lazy(self):
+        return self._liveness == 'lazy'
+
     @property
     def _mem_local(self):
         return self._space == 'local'
@@ -167,6 +191,10 @@ class Array(ArrayBasic):
     def _mem_mapped(self):
         return self._space == 'mapped'
 
+    @property
+    def _mem_host(self):
+        return self._space == 'host'
+
     @property
     def _mem_stack(self):
         return self._scope == 'stack'
@@ -186,8 +214,18 @@ class Array(ArrayBasic):
     def _make_pointer(self, dim):
         return PointerArray(name='p%s' % self.name, dimensions=dim, array=self)
 
-    # Pickling support
-    _pickle_kwargs = AbstractFunction._pickle_kwargs + ['dimensions', 'space', 'scope']
+
+class ArrayMapped(Array):
+
+    _C_structname = 'array'
+    _C_field_data = 'data'
+    _C_field_nbytes = 'nbytes'
+    _C_field_dmap = 'dmap'
+
+    _C_ctype = POINTER(type(_C_structname, (Structure,),
+                            {'_fields_': [(_C_field_data, c_restrict_void_p),
+                                          (_C_field_nbytes, c_ulong),
+                                          (_C_field_dmap, c_void_p)]}))
 
 
 class ArrayObject(ArrayBasic):
@@ -216,47 +254,31 @@ class ArrayObject(ArrayBasic):
 
     is_ObjectArray = True
 
+    __rkwargs__ = list(ArrayBasic.__rkwargs__) + ['dimensions', 'fields', 'pname']
+    __rkwargs__.remove('dtype')
+
     def __init_finalize__(self, *args, **kwargs):
         name = kwargs['name']
         fields = tuple(kwargs.pop('fields', ()))
 
         self._fields = fields
-        self._pname = "t%s" % name
+        self._pname = kwargs.pop('pname', 't%s' % name)
 
         super().__init_finalize__(*args, **kwargs)
 
     @classmethod
     def __dtype_setup__(cls, **kwargs):
-        return POINTER(type("t%s" % kwargs['name'],
-                            (Structure,),
-                            {'_fields_': cls.__pfields_setup__(**kwargs)}))
+        pname = kwargs.get('pname', 't%s' % kwargs['name'])
+        pfields = cls.__pfields_setup__(**kwargs)
+        return CtypesFactory.generate(pname, pfields)
 
     @classmethod
     def __pfields_setup__(cls, **kwargs):
         return [(i._C_name, i._C_ctype) for i in kwargs.get('fields', [])]
 
-    @cached_property
-    def _C_typename(self):
-        return ctypes_to_cstr(self.dtype)
-
-    @cached_property
-    def _C_typedata(self):
-        if self._is_composite_dtype:
-            return ctypes_to_cstr(self.dtype._type_)
-        else:
-            return self._C_typename
-
-    @cached_property
-    def _C_typedecl(self):
-        if self._is_composite_dtype:
-            return Struct(self.pname,
-                          [Value(ctypes_to_cstr(j), i) for i, j in self.pfields])
-        else:
-            return None
-
     @property
-    def _is_composite_dtype(self):
-        return len(self.fields) > 0
+    def _C_ctype(self):
+        return self.dtype
 
     @property
     def fields(self):
@@ -279,10 +301,6 @@ class ArrayObject(ArrayBasic):
     def _mem_stack(self):
         return True
 
-    # Pickling support
-    _pickle_kwargs = ArrayBasic._pickle_kwargs + ['dimensions', 'fields']
-    _pickle_kwargs.remove('dtype')
-
 
 class PointerArray(ArrayBasic):
 
@@ -306,6 +324,8 @@ class PointerArray(ArrayBasic):
 
     is_PointerArray = True
 
+    __rkwargs__ = ('name', 'dimensions', 'array')
+
     def __new__(cls, *args, **kwargs):
         kwargs.update({'options': {'evaluate': False}})
         return AbstractFunction.__new__(cls, *args, **kwargs)
@@ -321,8 +341,8 @@ class PointerArray(ArrayBasic):
         return kwargs['array'].dtype
 
     @property
-    def _C_typename(self):
-        return ctypes_to_cstr(POINTER(self._C_ctype))
+    def _C_ctype(self):
+        return POINTER(POINTER(dtype_to_ctype(self.dtype)))
 
     @property
     def dim(self):
@@ -332,6 +352,3 @@ class PointerArray(ArrayBasic):
     @property
     def array(self):
         return self._array
-
-    # Pickling support
-    _pickle_kwargs = ['name', 'dimensions', 'array']
