@@ -1,6 +1,6 @@
 import abc
 from collections import namedtuple
-from ctypes import POINTER, Structure, byref
+from ctypes import POINTER, _Pointer
 from functools import reduce
 from operator import mul
 
@@ -8,18 +8,16 @@ import numpy as np
 import sympy
 from sympy.core.assumptions import _assume_rules
 from cached_property import cached_property
-from cgen import Struct, Value
 
 from devito.data import default_allocator
-from devito.symbolics import aligned_indices
-from devito.tools import (Pickable, as_tuple, ctypes_to_cstr, dtype_to_cstr,
+from devito.tools import (Pickable, as_tuple, ctypes_to_cstr, ctypes_to_cgen,
                           dtype_to_ctype, frozendict, memoized_meth, sympy_mutex)
 from devito.types.args import ArgProvider
-from devito.types.caching import Cached
+from devito.types.caching import Cached, Uncached
 from devito.types.lazy import Evaluable
 from devito.types.utils import DimensionTuple
 
-__all__ = ['Symbol', 'Scalar', 'Indexed', 'Object', 'LocalObject', 'CompositeObject']
+__all__ = ['Symbol', 'Scalar', 'Indexed', 'IndexedData', 'DeviceMap']
 
 
 Size = namedtuple('Size', 'left right')
@@ -30,12 +28,39 @@ class CodeSymbol(object):
 
     """
     Abstract base class for objects representing symbols in the generated code.
+
+    The _C_* properties describe the object in C-land. For example its name and
+    its type.
+
+    The _mem_* properties describe the object memory allocation strategy. There
+    are three axes with a few possible values each:
+
+        *"liveness": `_mem_external`, `_mem_internal_eager`, `_mem_internal_lazy`
+        *"space": `_mem_local`, `_mem_mapped`, `_mem_host`
+        *"scope": `_mem_stack`, `_mem_heap`
+
+    For example, an object that is `<_mem_internal_lazy, _mem_local, _mem_heap>`
+    is allocated within the Operator entry point, on either the host or device
+    memory (but not both), and on the heap. Refer to the __doc__ of the single
+    _mem_* properties for more info.
     """
 
     @abc.abstractmethod
     def __init__(self, *args, **kwargs):
         return
 
+    @property
+    @abc.abstractmethod
+    def dtype(self):
+        """
+        The data type of the object in the generated code, represented as a
+        Python class:
+
+            * `numpy.dtype`: basic data types. For example, `np.float64 -> double`.
+            * `ctypes`: composite objects (e.g., structs), foreign types.
+        """
+        return
+
     @abc.abstractproperty
     def _C_name(self):
         """
@@ -47,8 +72,8 @@ class CodeSymbol(object):
         """
         return
 
-    @abc.abstractproperty
-    def _C_typename(self):
+    @property
+    def _C_typedata(self):
         """
         The type of the object in the generated code.
 
@@ -56,24 +81,39 @@ class CodeSymbol(object):
         -------
         str
         """
-        return
+        _type = self._C_ctype
+        while issubclass(_type, _Pointer):
+            _type = _type._type_
 
-    @abc.abstractproperty
-    def _C_typedata(self):
+        return ctypes_to_cstr(_type, qualifiers=self._C_typequals)
+
+    @property
+    def _C_typename(self):
         """
-        The type of the data values in the generated code.
+        The type used to carry around the object in the generated code.
+
+        If an object is expected to be passed by value, this will coincide
+        with `_C_typedata`.
+
+        Instead, if an object is passed via a one-dimensional pointer, then
+        `self._C_typename` will add a `*` to whatever type is returned by
+        `self._C_typedata`.
+
+        This can be customized at will by subclasses.
+
+        By default, `self._C_typename = self._C_typedata`.
 
         Returns
         -------
         str
         """
-        return
+        return ctypes_to_cstr(self._C_ctype, qualifiers=self._C_typequals)
 
     @abc.abstractproperty
     def _C_ctype(self):
         """
-        The type of the object as a ctypes object, necessary for jumping
-        from Python-land to generated-code-land.
+        The `_C_typename` of the object as a `ctypes` class, necessary for
+        jumping from Python-land to C-land.
 
         Returns
         -------
@@ -81,6 +121,18 @@ class CodeSymbol(object):
         """
         return
 
+    @cached_property
+    def _C_typequals(self):
+        """
+        The type qualifiers of the object in the generated code.
+        """
+        known_qualifiers = ('is_const', 'is_volatile')
+        qualifiers = []
+        for i in known_qualifiers:
+            if getattr(self, i, False):
+                qualifiers.append(i.split('_')[1])
+        return as_tuple(qualifiers)
+
     @property
     def _C_typedecl(self):
         """
@@ -92,7 +144,7 @@ class CodeSymbol(object):
             None if the type of the object can be expressed with a basic type,
             such as float or int, otherwise a cgen.Struct representing a C struct.
         """
-        return
+        return ctypes_to_cgen(self._C_ctype, fields=getattr(self, 'fields', None))
 
     @property
     def _C_symbol(self):
@@ -114,6 +166,83 @@ class CodeSymbol(object):
         """
         return (self,)
 
+    @property
+    def _mem_external(self):
+        """
+        True if the associated data is allocated and freed in Python, False otherwise.
+        """
+        return False
+
+    @property
+    def _mem_internal_eager(self):
+        """
+        True if the associated data is allocated and freed inside the first
+        Callable in which the symbol appears as a free variable.
+        """
+        return False
+
+    @property
+    def _mem_internal_lazy(self):
+        """
+        True if the associated data is allocated and freed at the level of
+        the Operator entry point.
+        """
+        return False
+
+    @property
+    def _mem_local(self):
+        """
+        True if the associated data is allocated in the underlying platform's
+        local memory space, False otherwise.
+
+        The local memory space is:
+
+            * the host DRAM if platform=CPU
+            * the device DRAM if platform=GPU
+        """
+        return False
+
+    @property
+    def _mem_mapped(self):
+        """
+        True if the associated data is allocated in the underlying platform's
+        local memory space and subsequently mapped to the underlying platform's
+        remote memory space, False otherwise.
+
+        The local memory space is:
+
+            * the host DRAM if platform=CPU
+            * the device DRAM if platform=GPU
+
+        The remote memory space is:
+
+            * the host DRAM if platform=GPU
+            * the device DRAM if platform=CPU
+        """
+        return False
+
+    @property
+    def _mem_host(self):
+        """
+        True if the associated data is systematically allocated in the host DRAM.
+        """
+        return False
+
+    @property
+    def _mem_stack(self):
+        """
+        True if the associated data should be allocated on the stack, False otherwise.
+        """
+        return False
+
+    @property
+    def _mem_heap(self):
+        """
+        True if the associated data was/is/will be allocated on the heap,
+        False otherwise.
+        """
+        return False
+
 
 class Basic(CodeSymbol):
 
@@ -151,7 +280,6 @@ class Basic(CodeSymbol):
 
     # Top hierarchy
     is_AbstractFunction = False
-    is_AbstractSymbol = False
     is_AbstractObject = False
 
     # Symbolic objects created internally by Devito
@@ -177,22 +305,22 @@ class Basic(CodeSymbol):
     is_TempFunction = False
     is_SparseTimeFunction = False
     is_SparseFunction = False
-    is_PrecomputedSparseFunction = False
-    is_PrecomputedSparseTimeFunction = False
 
     # Time dependence
     is_TimeDependent = False
 
-    # Tensor and Vector valued objects
-    is_VectorValued = False
-    is_TensorValued = False
-
-    # Basic symbolic object properties
-    is_Scalar = False
-
     # Some other properties
     is_PerfKnob = False  # Does it impact the Operator performance?
 
+    @property
+    def bound_symbols(self):
+        """
+        Unlike SymPy, we systematically define `bound_symbols` on all of
+        the API and internal objects that may be used to construct an
+        Operator.
+        """
+        return set()
+
 
 class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
 
@@ -231,6 +359,8 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
     is_imaginary = False
     is_commutative = True
 
+    __rkwargs__ = ('name', 'dtype', 'is_const')
+
     @classmethod
     def _filter_assumptions(cls, **kwargs):
         """Extract sympy.Symbol-specific kwargs."""
@@ -272,7 +402,6 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
 
     @property
     def dtype(self):
-        """The data type of the object."""
         return self._dtype
 
     @property
@@ -303,8 +432,7 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
     def function(self):
         return self
 
-    @property
-    def evaluate(self):
+    def _evaluate(self, **kwargs):
         return self
 
     def indexify(self):
@@ -322,13 +450,6 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
     def _C_name(self):
         return self.name
 
-    @property
-    def _C_typename(self):
-        return '%s%s' % ('const ' if self.is_const else '',
-                         dtype_to_cstr(self.dtype))
-
-    _C_typedata = _C_typename
-
     @property
     def _C_ctype(self):
         return dtype_to_ctype(self.dtype)
@@ -348,8 +469,6 @@ class AbstractSymbol(sympy.Symbol, Basic, Pickable, Evaluable):
         return self
 
     # Pickling support
-    _pickle_args = []
-    _pickle_kwargs = ['name', 'dtype', 'is_const']
     __reduce_ex__ = Pickable.__reduce_ex__
 
     def __getnewargs_ex__(self):
@@ -418,49 +537,27 @@ class Symbol(AbstractSymbol, Cached):
     __hash__ = Cached.__hash__
 
 
-class DataSymbol(AbstractSymbol, Cached):
+class DataSymbol(AbstractSymbol, Uncached):
 
     """
-    A scalar symbol, cached by both Devito and SymPy, which carries data.
+    A unique scalar symbol that carries data.
     """
 
-    @classmethod
-    def _cache_key(cls, *args, **kwargs):
-        return cls
-
     def __new__(cls, *args, **kwargs):
-        key = cls._cache_key(*args, **kwargs)
-        obj = cls._cache_get(key)
-
-        if obj is not None:
-            return obj
-
-        # Not in cache. Create a new Symbol via sympy.Symbol
+        # Create a new Symbol via sympy.Symbol
         name = kwargs.get('name') or args[0]
         assumptions, kwargs = cls._filter_assumptions(**kwargs)
 
-        # Create new, unique type instance from cls and the symbol name
-        newcls = type(name, (cls,), dict(cls.__dict__))
-
-        # Create the new Symbol and invoke __init__
-        newobj = sympy.Symbol.__new__(newcls, name, **assumptions)
+        # Note: use __xnew__ to bypass sympy caching
+        newobj = sympy.Symbol.__xnew__(cls, name, **assumptions)
 
         # Initialization
         newobj._dtype = cls.__dtype_setup__(**kwargs)
         newobj.__init_finalize__(*args, **kwargs)
 
-        # Store new instance in symbol cache
-        Cached.__init__(newobj, newcls)
-
         return newobj
 
-    __hash__ = Cached.__hash__
-
-    # Pickling support
-
-    @property
-    def _pickle_reconstruct(self):
-        return self.__class__.__base__
+    __hash__ = Uncached.__hash__
 
 
 class Scalar(Symbol, ArgProvider):
@@ -483,8 +580,6 @@ class Scalar(Symbol, ArgProvider):
         SymPy documentation for more information.
     """
 
-    is_Scalar = True
-
     @classmethod
     def __dtype_setup__(cls, **kwargs):
         return kwargs.get('dtype', np.float32)
@@ -722,6 +817,8 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
     is_imaginary = False
     is_commutative = True
 
+    __rkwargs__ = ('name', 'dtype', 'halo', 'padding')
+
     @classmethod
     def _cache_key(cls, *args, **kwargs):
         return cls, args
@@ -849,7 +946,7 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
 
     @property
     def indices(self):
-        """The indices (aka dimensions) of the object."""
+        """The indices of the object."""
         return DimensionTuple(*self.args, getters=self.dimensions)
 
     @property
@@ -882,22 +979,21 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         For example, if the original non-staggered function is f(x)
         then f(x) is on the grid and f(x + h_x/2) is off the grid.
         """
-        return self._check_indices(inds=self.indices)
+        for i, j, d in zip(self.indices, self.indices_ref, self.dimensions):
+            # Two indices are aligned if they differ by an Integer*spacing.
+            v = i - j
+            try:
+                if int(v/d.spacing) != v/d.spacing:
+                    return False
+            except TypeError:
+                return False
+        return True
 
-    @memoized_meth
-    def _check_indices(self, inds=None):
-        """
-        Check if the function indices are aligned with the dimensions.
-        """
-        inds = inds or self.indices
-        return all([aligned_indices(i, j, d.spacing) for i, j, d in
-                    zip(inds, self.indices_ref, self.dimensions)])
-
-    @property
-    def evaluate(self):
+    def _evaluate(self, **kwargs):
         # Average values if at a location not on the Function's grid
         if self._is_on_grid:
             return self
+
         weight = 1.0
         avg_list = [self]
         is_averaged = False
@@ -922,7 +1018,6 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
 
     @property
     def dtype(self):
-        """The data type of the object."""
         return self._dtype
 
     @property
@@ -944,72 +1039,28 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
                     for i, j, k in zip(domain, halo, padding))
         return DimensionTuple(*ret, getters=self.dimensions)
 
+    @cached_property
+    def bound_symbols(self):
+        return ({self._C_symbol, self.indexed} |
+                set().union(*[d.bound_symbols for d in self.dimensions]))
+
     @cached_property
     def indexed(self):
         """The wrapped IndexedData object."""
         return IndexedData(self.name, shape=self.shape, function=self.function)
 
-    @property
-    def _mem_local(self):
+    @cached_property
+    def dmap(self):
         """
-        True if the associated data is allocated in the underlying platform's
-        local memory space, False otherwise.
-
-        The local memory space is:
-
-            * the host DRAM if platform=CPU
-            * the device DRAM if platform=GPU
-
-        Defaults to False because AbstractFunctions are normally allocated in
-        the host DRAM regardless of the underlying platform, and then
-        dynamically mapped to device DRAM (directly from the generated code) if
-        the underlying platform turns out to be a device.
+        A symbolic pointer to the device data map. If there's no such a map,
+        return None.
         """
-        return False
-
-    @property
-    def _mem_mapped(self):
-        """
-        True if the associated data is allocated in the underlying platform's
-        local memory space and subsequently mapped to the underlying platform's
-        remote memory space, False otherwise.
-
-        The local memory space is:
-
-            * the host DRAM if platform=CPU
-            * the device DRAM if platform=GPU
-
-        The remote memory space is:
-
-            * the host DRAM if platform=GPU
-            * the device DRAM if platform=CPU
-
-        Defaults to True, thus relaxing `_mem_local`.
-        """
-        return not self._mem_local
-
-    @property
-    def _mem_external(self):
-        """
-        True if the associated data was/is/will be allocated directly
-        from Python (e.g., via NumPy arrays), False otherwise.
-        """
-        return False
-
-    @property
-    def _mem_stack(self):
-        """
-        True if the associated data should be allocated on the stack, False otherwise.
-        """
-        return False
-
-    @property
-    def _mem_heap(self):
-        """
-        True if the associated data was/is/will be allocated on the heap,
-        False otherwise.
-        """
-        return False
+        if self._mem_mapped:
+            return DeviceMap('d_%s' % self.name, shape=self.shape, function=self.function)
+        elif self._mem_local:
+            return self.indexed
+        else:
+            return None
 
     @property
     def size(self):
@@ -1040,14 +1091,14 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
     def _C_name(self):
         return "%s_vec" % self.name
 
-    @property
-    def _C_typedata(self):
-        return dtype_to_cstr(self.dtype)
-
     @cached_property
     def _C_symbol(self):
         return BoundSymbol(name=self._C_name, dtype=self.dtype, function=self.function)
 
+    @property
+    def _mem_mapped(self):
+        return not self._mem_local
+
     def _make_pointer(self):
         """Generate a symbolic pointer to self."""
         raise NotImplementedError
@@ -1158,201 +1209,29 @@ class AbstractFunction(sympy.Function, Basic, Cached, Pickable, Evaluable):
         return self.indexed[index]
 
     # Pickling support
-    _pickle_kwargs = ['name', 'dtype', 'halo', 'padding']
     __reduce_ex__ = Pickable.__reduce_ex__
 
+    # Reconstruction support
     @property
-    def _pickle_reconstruct(self):
+    def _rcls(self):
         return self.__class__.__base__
 
 
-class AbstractObject(Basic, sympy.Basic, Pickable):
-
-    """
-    Base class for pointers to objects with derived type.
-
-    The hierarchy is structured as follows
-
-                         AbstractObject
-                                |
-                 ---------------------------------
-                 |                               |
-              Object                       LocalObject
-                 |
-          CompositeObject
-
-    Warnings
-    --------
-    AbstractObjects are created and managed directly by Devito.
-    """
-
-    is_AbstractObject = True
-
-    def __new__(cls, *args, **kwargs):
-        with sympy_mutex:
-            obj = sympy.Basic.__new__(cls)
-        obj.__init__(*args, **kwargs)
-        return obj
-
-    def __init__(self, name, dtype):
-        self.name = name
-        self.dtype = dtype
-
-    def __repr__(self):
-        return self.name
-
-    __str__ = __repr__
-
-    def _hashable_content(self):
-        return (self.name, self.dtype)
-
-    @property
-    def free_symbols(self):
-        return {self}
-
-    @property
-    def _C_name(self):
-        return self.name
-
-    @property
-    def _C_typename(self):
-        return ctypes_to_cstr(self.dtype)
-
-    @property
-    def _C_typedata(self):
-        return self._C_typename
-
-    @property
-    def _C_ctype(self):
-        return self.dtype
-
-    @property
-    def function(self):
-        return self
-
-    # Pickling support
-    _pickle_args = ['name', 'dtype']
-    __reduce_ex__ = Pickable.__reduce_ex__
-
-
-class Object(AbstractObject, ArgProvider):
-
-    """
-    Pointer to object with derived type, provided by an outer scope.
-    """
-
-    is_Object = True
-
-    def __init__(self, name, dtype, value=None):
-        super(Object, self).__init__(name, dtype)
-        self.value = value
-
-    @property
-    def _arg_names(self):
-        return (self.name,)
-
-    def _arg_defaults(self):
-        if callable(self.value):
-            return {self.name: self.value()}
-        else:
-            return {self.name: self.value}
-
-    def _arg_values(self, **kwargs):
-        """
-        Produce runtime values for this Object after evaluating user input.
-
-        Parameters
-        ----------
-        **kwargs
-            Dictionary of user-provided argument overrides.
-        """
-        if self.name in kwargs:
-            return {self.name: kwargs.pop(self.name)}
-        else:
-            return self._arg_defaults()
-
-
-class CompositeObject(Object):
-
-    """
-    Pointer to object with composite type (e.g., a C struct), provided
-    by an outer scope.
-    """
-
-    _dtype_cache = {}
-
-    @classmethod
-    def _generate_unique_dtype(cls, pname, pfields):
-        dtype = POINTER(type(pname, (Structure,), {'_fields_': pfields}))
-        key = (pname, tuple(pfields))
-        return cls._dtype_cache.setdefault(key, dtype)
-
-    def __init__(self, name, pname, pfields, value=None):
-        dtype = CompositeObject._generate_unique_dtype(pname, pfields)
-        value = self.__value_setup__(dtype, value)
-        super(CompositeObject, self).__init__(name, dtype, value)
-
-    def __value_setup__(self, dtype, value):
-        return value or byref(dtype._type_())
-
-    @property
-    def pfields(self):
-        return tuple(self.dtype._type_._fields_)
-
-    @property
-    def pname(self):
-        return self.dtype._type_.__name__
-
-    @property
-    def fields(self):
-        return [i for i, _ in self.pfields]
-
-    def _hashable_content(self):
-        return (self.name, self.pfields)
-
-    @cached_property
-    def _C_typedecl(self):
-        return Struct(self.pname, [Value(ctypes_to_cstr(j), i) for i, j in self.pfields])
-
-    # Pickling support
-    _pickle_args = ['name', 'pname', 'pfields']
-    _pickle_kwargs = []
-
-
-class LocalObject(AbstractObject):
-
-    """
-    Pointer to object with derived type, defined in the local scope.
-    """
-
-    is_LocalObject = True
-
-    dtype = None
-    """
-    LocalObjects encode their dtype as a class attribute.
-    """
-
-    def __init__(self, name, constructor_args=None, **kwargs):
-        self.name = name
-        self.constructor_args = as_tuple(constructor_args)
-
-    # Pickling support
-    _pickle_args = ['name']
-    _pickle_kwargs = ['constructor_args']
-
-
 # Extended SymPy hierarchy follows, for essentially two reasons:
 # - To keep track of `function`
 # - To override SymPy caching behaviour
 
 
-class IndexedData(sympy.IndexedBase, Pickable, CodeSymbol):
+class IndexedBase(sympy.IndexedBase, Basic, Pickable):
 
     """
     Wrapper class that inserts a pointer to the symbolic data object.
     """
 
-    def __new__(cls, label, shape=None, function=None):
+    __rargs__ = ('label', 'shape')
+    __rkwargs__ = ('function',)
+
+    def __new__(cls, label, shape, function=None):
         # Make sure `label` is a devito.Symbol, not a sympy.Symbol
         if isinstance(label, str):
             label = Symbol(name=label, dtype=None)
@@ -1361,45 +1240,61 @@ class IndexedData(sympy.IndexedBase, Pickable, CodeSymbol):
         obj.function = function
         return obj
 
-    def func(self, *args):
-        obj = super(IndexedData, self).func(*args)
-        obj.function = self.function
-        return obj
+    func = Pickable._rebuild
 
     def __getitem__(self, indices, **kwargs):
         """Produce a types.Indexed, rather than a sympy.Indexed."""
-        indexed = super(IndexedData, self).__getitem__(indices, **kwargs)
-        return Indexed(*indexed.args)
+        return Indexed(self, *as_tuple(indices))
+
+    def _hashable_content(self):
+        return super()._hashable_content() + (self.function,)
 
     @property
     def _C_name(self):
         return self.name
 
-    @cached_property
-    def _C_typename(self):
-        return ctypes_to_cstr(self._C_ctype)
-
-    @cached_property
-    def _C_typedata(self):
-        return dtype_to_cstr(self.dtype)
-
     @cached_property
     def _C_ctype(self):
-        return POINTER(dtype_to_ctype(self.dtype))
+        try:
+            return POINTER(dtype_to_ctype(self.dtype))
+        except TypeError:
+            # `dtype` is a ctypes-derived type!
+            return self.dtype
 
     @property
     def base(self):
         return self
 
+    @property
+    def indices(self):
+        return ()
+
     @property
     def dtype(self):
         return self.function.dtype
 
+    @cached_property
+    def free_symbols(self):
+        ret = {self}
+        for i in self.indices:
+            try:
+                ret.update(i.free_symbols)
+            except AttributeError:
+                pass
+        return ret
+
     # Pickling support
-    _pickle_kwargs = ['label', 'shape', 'function']
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
+class IndexedData(IndexedBase):
+    pass
+
+
+class DeviceMap(IndexedBase):
+    pass
+
+
 class BoundSymbol(AbstractSymbol):
 
     """
@@ -1424,6 +1319,13 @@ class BoundSymbol(AbstractSymbol):
     def function(self):
         return self._function
 
+    def _hashable_content(self):
+        return super()._hashable_content() + (self.function,)
+
+    @property
+    def _C_ctype(self):
+        return self.function._C_ctype
+
 
 class Indexed(sympy.Indexed):
 
