@@ -42,19 +42,19 @@ class Constant(DataSymbol, ArgProvider):
 
     is_Input = True
     is_Constant = True
-    is_Scalar = True
+
+    __rkwargs__ = DataSymbol.__rkwargs__ + ('value',)
 
     def __init_finalize__(self, *args, **kwargs):
-        self._value = kwargs.get('value', 0)
+        self._value = kwargs.pop('value', 0)
+
+        kwargs.setdefault('is_const', True)
+        super().__init_finalize__(*args, **kwargs)
 
     @classmethod
     def __dtype_setup__(cls, **kwargs):
         return kwargs.get('dtype', np.float32)
 
-    @property
-    def is_const(self):
-        return True
-
     @property
     def value(self):
         return self.data
@@ -111,5 +111,3 @@ class Constant(DataSymbol, ArgProvider):
                         "Constant data type %s" % (key.dtype, self.name, self.dtype))
         except AttributeError:
             pass
-
-    _pickle_kwargs = DataSymbol._pickle_kwargs + ['value']
