@@ -1,5 +1,5 @@
 from collections import namedtuple
-from ctypes import POINTER, Structure, c_void_p, c_int, cast, byref
+from ctypes import POINTER, Structure, c_int, c_ulong, c_void_p, cast, byref
 from functools import wraps, reduce
 from math import ceil
 from operator import mul
@@ -8,7 +8,6 @@ import numpy as np
 import sympy
 from psutil import virtual_memory
 from cached_property import cached_property
-from cgen import Struct, Value
 
 from devito.builtins import assign
 from devito.data import (DOMAIN, OWNED, HALO, NOPAD, FULL, LEFT, CENTER, RIGHT,
@@ -19,8 +18,8 @@ from devito.mpi import MPI
 from devito.parameters import configuration
 from devito.symbolics import FieldFromPointer
 from devito.finite_differences import Differentiable, generate_fd_shortcuts
-from devito.tools import (ReducerMap, as_tuple, flatten, is_integer,
-                          ctypes_to_cstr, memoized_meth, dtype_to_ctype)
+from devito.tools import (ReducerMap, as_tuple, c_restrict_void_p, flatten, is_integer,
+                          memoized_meth, dtype_to_ctype, humanbytes)
 from devito.types.dimension import Dimension
 from devito.types.args import ArgProvider
 from devito.types.caching import CacheManager
@@ -58,6 +57,8 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
     The type of the underlying data object.
     """
 
+    __rkwargs__ = AbstractFunction.__rkwargs__ + ('grid', 'staggered', 'initializer')
+
     def __init_finalize__(self, *args, **kwargs):
         # A `Distributor` to handle domain decomposition (only relevant for MPI)
         self._distributor = self.__distributor_setup__(**kwargs)
@@ -116,7 +117,8 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         @wraps(func)
         def wrapper(self):
             if self._data is None:
-                debug("Allocating memory for %s%s" % (self.name, self.shape_allocated))
+                debug("Allocating host memory for %s%s [%s]"
+                      % (self.name, self.shape_allocated, humanbytes(self.nbytes)))
 
                 # Clear up both SymPy and Devito caches to drop unreachable data
                 CacheManager.clear(force=False)
@@ -306,6 +308,10 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
             retval.append(size.glb if size is not None else s)
         return tuple(retval)
 
+    @property
+    def symbolic_shape(self):
+        return tuple(self._C_get_field(FULL, d).size for d in self.dimensions)
+
     @property
     def size_global(self):
         """
@@ -565,8 +571,11 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         self._is_halo_dirty = True
         offset = getattr(getattr(self, '_offset_%s' % region.name)[dim], side.name)
         size = getattr(getattr(self, '_size_%s' % region.name)[dim], side.name)
-        index_array = [slice(offset, offset+size) if d is dim else slice(None)
-                       for d in self.dimensions]
+        index_array = [
+            slice(offset, offset+size) if d is dim else slice(pl, s - pr)
+            for d, s, (pl, pr)
+            in zip(self.dimensions, self.shape_allocated, self._padding)
+        ]
         return np.asarray(self._data[index_array])
 
     @property
@@ -654,7 +663,6 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
             return self._initializer
 
     _C_structname = 'dataobj'
-    _C_typename = 'struct %s *' % _C_structname
     _C_field_data = 'data'
     _C_field_size = 'size'
     _C_field_nopad_size = 'npsize'
@@ -662,24 +670,17 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
     _C_field_halo_size = 'hsize'
     _C_field_halo_ofs = 'hofs'
     _C_field_owned_ofs = 'oofs'
-
-    _C_typedecl = Struct(_C_structname,
-                         [Value('%srestrict' % ctypes_to_cstr(c_void_p), _C_field_data),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_size),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_nopad_size),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_domain_size),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_halo_size),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_halo_ofs),
-                          Value(ctypes_to_cstr(POINTER(c_int)), _C_field_owned_ofs)])
+    _C_field_dmap = 'dmap'
 
     _C_ctype = POINTER(type(_C_structname, (Structure,),
-                            {'_fields_': [(_C_field_data, c_void_p),
-                                          (_C_field_size, POINTER(c_int)),
-                                          (_C_field_nopad_size, POINTER(c_int)),
-                                          (_C_field_domain_size, POINTER(c_int)),
+                            {'_fields_': [(_C_field_data, c_restrict_void_p),
+                                          (_C_field_size, POINTER(c_ulong)),
+                                          (_C_field_nopad_size, POINTER(c_ulong)),
+                                          (_C_field_domain_size, POINTER(c_ulong)),
                                           (_C_field_halo_size, POINTER(c_int)),
                                           (_C_field_halo_ofs, POINTER(c_int)),
-                                          (_C_field_owned_ofs, POINTER(c_int))]}))
+                                          (_C_field_owned_ofs, POINTER(c_int)),
+                                          (_C_field_dmap, c_void_p)]}))
 
     def _C_make_dataobj(self, data):
         """
@@ -687,16 +688,19 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         an Operator.
         """
         dataobj = byref(self._C_ctype._type_())
-        dataobj._obj.data = data.ctypes.data_as(c_void_p)
-        dataobj._obj.size = (c_int*self.ndim)(*data.shape)
+        dataobj._obj.data = data.ctypes.data_as(c_restrict_void_p)
+        dataobj._obj.size = (c_ulong*self.ndim)(*data.shape)
         # MPI-related fields
-        dataobj._obj.npsize = (c_int*self.ndim)(*[i - sum(j) for i, j in
-                                                  zip(data.shape, self._size_padding)])
-        dataobj._obj.dsize = (c_int*self.ndim)(*self._size_domain)
+        dataobj._obj.npsize = (c_ulong*self.ndim)(*[i - sum(j) for i, j in
+                                                    zip(data.shape, self._size_padding)])
+        dataobj._obj.dsize = (c_ulong*self.ndim)(*self._size_domain)
         dataobj._obj.hsize = (c_int*(self.ndim*2))(*flatten(self._size_halo))
         dataobj._obj.hofs = (c_int*(self.ndim*2))(*flatten(self._offset_halo))
         dataobj._obj.oofs = (c_int*(self.ndim*2))(*flatten(self._offset_owned))
 
+        # Fields used only within C-land
+        dataobj._obj.dmap = c_void_p(0)
+
         # stash a reference to the array on _obj, so we don't let it get freed
         # while we hold onto _obj
         dataobj._obj.underlying_array = data
@@ -873,10 +877,6 @@ class DiscreteFunction(AbstractFunction, ArgProvider, Differentiable):
         key = alias or self
         return {key.name: self._C_make_dataobj(args[key.name])}
 
-    # Pickling support
-    _pickle_kwargs = AbstractFunction._pickle_kwargs +\
-        ['grid', 'staggered', 'initializer']
-
 
 class Function(DiscreteFunction):
 
@@ -977,6 +977,9 @@ class Function(DiscreteFunction):
 
     is_Function = True
 
+    __rkwargs__ = (DiscreteFunction.__rkwargs__ +
+                   ('space_order', 'shape_global', 'dimensions'))
+
     def _cache_meta(self):
         # Attach additional metadata to self's cache entry
         return {'nbytes': self.size}
@@ -1182,10 +1185,6 @@ class Function(DiscreteFunction):
         tot = self.sum(p, dims)
         return tot / len(tot.args)
 
-    # Pickling support
-    _pickle_kwargs = DiscreteFunction._pickle_kwargs +\
-        ['space_order', 'shape_global', 'dimensions']
-
 
 class TimeFunction(Function):
 
@@ -1305,6 +1304,8 @@ class TimeFunction(Function):
     _time_position = 0
     """Position of time index among the function indices."""
 
+    __rkwargs__ = Function.__rkwargs__ + ('time_order', 'save', 'time_dim')
+
     def __init_finalize__(self, *args, **kwargs):
         self.time_dim = kwargs.get('time_dim', self.dimensions[self._time_position])
         self._time_order = kwargs.get('time_order', 1)
@@ -1407,6 +1408,10 @@ class TimeFunction(Function):
     def _time_size(self):
         return self.shape_allocated[self._time_position]
 
+    @property
+    def time_size(self):
+        return self._time_size
+
     @property
     def _time_buffering(self):
         return not is_integer(self.save)
@@ -1423,9 +1428,6 @@ class TimeFunction(Function):
                                   "value `%s`, found `%d` instead"
                                   % (self._time_size, self.name, key_time_size))
 
-    # Pickling support
-    _pickle_kwargs = Function._pickle_kwargs + ['time_order', 'save', 'time_dim']
-
 
 class SubFunction(Function):
 
@@ -1436,6 +1438,8 @@ class SubFunction(Function):
     parent DiscreteFunction.
     """
 
+    __rkwargs__ = Function.__rkwargs__ + ('parent',)
+
     def __init_finalize__(self, *args, **kwargs):
         super(SubFunction, self).__init_finalize__(*args, **kwargs)
         self._parent = kwargs['parent']
@@ -1458,8 +1462,6 @@ class SubFunction(Function):
     def parent(self):
         return self._parent
 
-    _pickle_kwargs = Function._pickle_kwargs + ['parent']
-
 
 class TempFunction(DiscreteFunction):
 
@@ -1494,6 +1496,8 @@ class TempFunction(DiscreteFunction):
 
     is_TempFunction = True
 
+    __rkwargs__ = DiscreteFunction.__rkwargs__ + ('dimensions', 'pointer_dim')
+
     def __init_finalize__(self, *args, **kwargs):
         super().__init_finalize__(*args, **kwargs)
 
@@ -1553,7 +1557,8 @@ class TempFunction(DiscreteFunction):
         ret = tuple(sum(i) for i in zip(domain, halo))
         return DimensionTuple(*ret, getters=self.dimensions)
 
-    shape_allocated = DiscreteFunction.symbolic_shape
+    shape_allocated = AbstractFunction.symbolic_shape
+    symbolic_shape = AbstractFunction.symbolic_shape
 
     def make(self, shape=None, initializer=None, allocator=None, **kwargs):
         """
@@ -1614,9 +1619,6 @@ class TempFunction(DiscreteFunction):
         else:
             raise InvalidArgument("TempFunction `%s` lacks override" % self.name)
 
-    # Pickling support
-    _pickle_kwargs = DiscreteFunction._pickle_kwargs + ['dimensions', 'pointer_dim']
-
 
 class AliasFunction(DiscreteFunction):
 
@@ -1630,11 +1632,23 @@ class AliasFunction(DiscreteFunction):
     """
 
     __indices_setup__ = Function.__indices_setup__
-    __shape_setup__ = Function.__shape_setup__
 
-    @property
-    def _mem_mapped(self):
-        return False
+    def __init_finalize__(self, *args, **kwargs):
+        self.save = kwargs.pop('save', None)
+
+        super().__init_finalize__(*args, **kwargs)
+
+    @classmethod
+    def __shape_setup__(cls, **kwargs):
+        # We don't care about the actual shape, nobody will ever query it, and
+        # bypassing the `__shape_setup__` makes it easier to use AliasFunction
+        # to map various DiscreteFunction subclasses
+        return None
+
+    def __distributor_setup__(self, **kwargs):
+        # Same rational as `__shape_setup__`  -- we don't care about the
+        # underlying distributor
+        return None
 
     @property
     def data(self):
