@@ -7,13 +7,14 @@ from cached_property import cached_property
 from devito.data import LEFT, RIGHT
 from devito.exceptions import InvalidArgument
 from devito.logger import debug
-from devito.tools import Pickable, dtype_to_cstr, is_integer
+from devito.tools import Pickable, is_integer
 from devito.types.args import ArgProvider
 from devito.types.basic import Symbol, DataSymbol, Scalar
 
 __all__ = ['Dimension', 'SpaceDimension', 'TimeDimension', 'DefaultDimension',
            'CustomDimension', 'SteppingDimension', 'SubDimension', 'ConditionalDimension',
-           'dimensions', 'ModuloDimension', 'IncrDimension', 'BlockDimension']
+           'ModuloDimension', 'IncrDimension', 'BlockDimension', 'StencilDimension',
+           'dimensions']
 
 
 Thickness = namedtuple('Thickness', 'left right')
@@ -104,8 +105,8 @@ class Dimension(ArgProvider):
     is_Incr = False
     is_Block = False
 
-    _C_typename = 'const %s' % dtype_to_cstr(np.int32)
-    _C_typedata = _C_typename
+    __rargs__ = ('name',)
+    __rkwargs__ = ('spacing',)
 
     def __new__(cls, *args, **kwargs):
         """
@@ -183,25 +184,21 @@ class Dimension(ArgProvider):
     def root(self):
         return self
 
+    @cached_property
+    def bound_symbols(self):
+        candidates = [self.symbolic_min, self.symbolic_max, self.symbolic_size,
+                      self.symbolic_incr]
+        return frozenset(i for i in candidates if not i.is_Number)
+
     @property
     def _maybe_distributed(self):
         """Could it be a distributed Dimension?"""
         return True
 
-    @property
-    def _C_name(self):
-        return self.name
-
     @cached_property
     def _defines(self):
         return frozenset({self})
 
-    @cached_property
-    def _defines_symbols(self):
-        candidates = [self.symbolic_min, self.symbolic_max, self.symbolic_size,
-                      self.symbolic_incr]
-        return frozenset(i for i in candidates if not i.is_Number)
-
     @property
     def _arg_names(self):
         """Tuple of argument names introduced by the Dimension."""
@@ -307,8 +304,6 @@ class Dimension(ArgProvider):
                   self.max_name, args[self.max_name], self.name)
 
     # Pickling support
-    _pickle_args = ['name']
-    _pickle_kwargs = ['spacing']
     __reduce_ex__ = Pickable.__reduce_ex__
     __getnewargs_ex__ = Pickable.__getnewargs_ex__
 
@@ -320,7 +315,7 @@ class BasicDimension(Dimension, Symbol):
     def __new__(cls, *args, **kwargs):
         return Symbol.__new__(cls, *args, **kwargs)
 
-    def __init_finalize__(self, name, spacing=None):
+    def __init_finalize__(self, name, spacing=None, **kwargs):
         self._spacing = spacing or Scalar(name='h_%s' % name, is_const=True)
 
 
@@ -349,7 +344,7 @@ class DefaultDimension(Dimension, DataSymbol):
     def __new__(cls, *args, **kwargs):
         return DataSymbol.__new__(cls, *args, **kwargs)
 
-    def __init_finalize__(self, name, spacing=None, default_value=None):
+    def __init_finalize__(self, name, spacing=None, default_value=None, **kwargs):
         self._spacing = spacing or Scalar(name='h_%s' % name, is_const=True)
         self._default_value = default_value or 0
 
@@ -422,6 +417,9 @@ class DerivedDimension(BasicDimension):
 
     is_Derived = True
 
+    __rargs__ = Dimension.__rargs__ + ('parent',)
+    __rkwargs__ = ()
+
     def __init_finalize__(self, name, parent):
         assert isinstance(parent, Dimension)
         self._parent = parent
@@ -453,10 +451,6 @@ class DerivedDimension(BasicDimension):
         """A DerivedDimension performs no runtime checks."""
         return
 
-    # Pickling support
-    _pickle_args = Dimension._pickle_args + ['parent']
-    _pickle_kwargs = []
-
 
 # ***
 # The Dimensions below are exposed in the user API. They can only be created by
@@ -513,7 +507,11 @@ class SubDimension(DerivedDimension):
 
     is_Sub = True
 
-    def __init_finalize__(self, name, parent, left, right, thickness, local):
+    __rargs__ = (DerivedDimension.__rargs__ +
+                 ('symbolic_min', 'symbolic_max', 'thickness', 'local'))
+    __rkwargs__ = ()
+
+    def __init_finalize__(self, name, parent, left, right, thickness, local, **kwargs):
         super().__init_finalize__(name, parent)
         self._interval = sympy.Interval(left, right)
         self._thickness = Thickness(*thickness)
@@ -574,6 +572,11 @@ class SubDimension(DerivedDimension):
     def thickness(self):
         return self._thickness
 
+    @cached_property
+    def bound_symbols(self):
+        # Add thickness symbols
+        return frozenset().union(*[i.free_symbols for i in super().bound_symbols])
+
     @property
     def _maybe_distributed(self):
         return not self.local
@@ -582,10 +585,6 @@ class SubDimension(DerivedDimension):
     def _thickness_map(self):
         return dict(self.thickness)
 
-    @cached_property
-    def _defines_symbols(self):
-        return super()._defines_symbols | frozenset(self._thickness_map)
-
     @cached_property
     def _offset_left(self):
         # The left extreme of the SubDimension can be related to either the
@@ -671,11 +670,6 @@ class SubDimension(DerivedDimension):
 
         return {i.name: v for i, v in zip(self._thickness_map, (ltkn, rtkn))}
 
-    # Pickling support
-    _pickle_args = DerivedDimension._pickle_args +\
-        ['symbolic_min', 'symbolic_max', 'thickness', 'local']
-    _pickle_kwargs = []
-
 
 class ConditionalDimension(DerivedDimension):
 
@@ -688,7 +682,7 @@ class ConditionalDimension(DerivedDimension):
     ----------
     name : str
         Name of the dimension.
-    parent : Dimension
+    parent : Dimension, optional
         The parent Dimension.
     factor : int, optional
         The number of iterations between two executions of the if-branch. If None
@@ -753,9 +747,17 @@ class ConditionalDimension(DerivedDimension):
     is_NonlinearDerived = True
     is_Conditional = True
 
-    def __init_finalize__(self, name, parent, factor=None, condition=None,
-                          indirect=False):
+    __rkwargs__ = DerivedDimension.__rkwargs__ + ('factor', 'condition', 'indirect')
+
+    def __init_finalize__(self, name, parent=None, factor=None, condition=None,
+                          indirect=False, **kwargs):
+        # `parent=None` degenerates to a ConditionalDimension outside of
+        # any iteration space
+        if parent is None:
+            parent = BOTTOM
+
         super().__init_finalize__(name, parent)
+
         self._factor = factor
         self._condition = condition
         self._indirect = indirect
@@ -791,9 +793,6 @@ class ConditionalDimension(DerivedDimension):
             pass
         return retval
 
-    # Pickling support
-    _pickle_kwargs = DerivedDimension._pickle_kwargs + ['factor', 'condition', 'indirect']
-
 
 # ***
 # The Dimensions below are for internal use only. They are created by the compiler
@@ -842,8 +841,10 @@ class ModuloDimension(DerivedDimension):
     is_Modulo = True
     is_SubIterator = True
 
-    def __init_finalize__(self, name, parent,
-                          offset=None, modulo=None, incr=None, origin=None):
+    __rkwargs__ = ('offset', 'modulo', 'incr', 'origin')
+
+    def __init_finalize__(self, name, parent, offset=None, modulo=None, incr=None,
+                          origin=None, **kwargs):
         super().__init_finalize__(name, parent)
 
         # Sanity check
@@ -907,17 +908,14 @@ class ModuloDimension(DerivedDimension):
         except (TypeError, ValueError):
             return incr
 
+    @cached_property
+    def bound_symbols(self):
+        return set(self.parent.bound_symbols)
+
     def _arg_defaults(self, **kwargs):
-        """
-        A ModuloDimension provides no arguments, so this method returns an empty dict.
-        """
         return {}
 
     def _arg_values(self, *args, **kwargs):
-        """
-        A ModuloDimension provides no arguments, so there are no argument values
-        to be derived.
-        """
         return {}
 
     # Override SymPy arithmetic operators to exploit properties of modular arithmetic
@@ -942,9 +940,6 @@ class ModuloDimension(DerivedDimension):
             pass
         return super().__sub__(other)
 
-    # Pickling support
-    _pickle_kwargs = ['offset', 'modulo', 'incr', 'origin']
-
 
 class AbstractIncrDimension(DerivedDimension):
 
@@ -977,7 +972,10 @@ class AbstractIncrDimension(DerivedDimension):
 
     is_Incr = True
 
-    def __init_finalize__(self, name, parent, _min, _max, step=None, size=None):
+    __rargs__ = ('name', 'parent', 'symbolic_min', 'symbolic_max')
+    __rkwargs__ = ('step', 'size')
+
+    def __init_finalize__(self, name, parent, _min, _max, step=None, size=None, **kwargs):
         super().__init_finalize__(name, parent)
         self._min = _min
         self._max = _max
@@ -1040,9 +1038,12 @@ class AbstractIncrDimension(DerivedDimension):
         except (TypeError, ValueError):
             return self.step
 
-    # Pickling support
-    _pickle_args = ['name', 'parent', 'symbolic_min', 'symbolic_max']
-    _pickle_kwargs = ['step', 'size']
+    @cached_property
+    def bound_symbols(self):
+        ret = set(self.parent.bound_symbols)
+        if self.symbolic_incr.is_Symbol:
+            ret.add(self.symbolic_incr)
+        return frozenset(ret)
 
 
 class IncrDimension(AbstractIncrDimension):
@@ -1058,6 +1059,80 @@ class IncrDimension(AbstractIncrDimension):
     is_SubIterator = True
 
 
+class BlockDimension(AbstractIncrDimension):
+
+    """
+    Dimension symbol for lowering TILABLE Dimensions.
+    """
+
+    is_Block = True
+    is_PerfKnob = True
+
+    @cached_property
+    def _arg_names(self):
+        try:
+            return (self.step.name,)
+        except AttributeError:
+            # `step` not a Symbol
+            return ()
+
+    def _arg_defaults(self, **kwargs):
+        # TODO: need a heuristic to pick a default incr size
+        # TODO: move default value to __new__
+        try:
+            return {self.step.name: 8}
+        except AttributeError:
+            # `step` not a Symbol
+            return {}
+
+    def _arg_values(self, interval, grid, args=None, **kwargs):
+        try:
+            name = self.step.name
+        except AttributeError:
+            # `step` not a Symbol
+            return {}
+
+        if name in kwargs:
+            return {name: kwargs.pop(name)}
+        elif isinstance(self.parent, BlockDimension):
+            # `self` is a BlockDimension within an outer BlockDimension, but
+            # no value supplied -> the sub-block will span the entire block
+            return {name: args[self.parent.step.name]}
+        else:
+            value = self._arg_defaults()[name]
+            if value <= args[self.root.max_name] - args[self.root.min_name] + 1:
+                return {name: value}
+            else:
+                # Avoid OOB (will end up here only in case of tiny iteration spaces)
+                return {name: 1}
+
+    def _arg_check(self, args, *_args):
+        try:
+            name = self.step.name
+        except AttributeError:
+            # `step` not a Symbol
+            return
+
+        value = args[name]
+        if isinstance(self.parent, BlockDimension):
+            # sub-BlockDimensions must be perfect divisors of their parent
+            parent_value = args[self.parent.step.name]
+            if parent_value % value > 0:
+                raise InvalidArgument("Illegal block size `%s=%d`: sub-block sizes "
+                                      "must divide the parent block size evenly (`%s=%d`)"
+                                      % (name, value, self.parent.step.name,
+                                         parent_value))
+        else:
+            if value < 0:
+                raise InvalidArgument("Illegal block size `%s=%d`: it should be > 0"
+                                      % (name, value))
+            if value > args[self.root.max_name] - args[self.root.min_name] + 1:
+                # Avoid OOB
+                raise InvalidArgument("Illegal block size `%s=%d`: it's greater than the "
+                                      "iteration range and it will cause an OOB access"
+                                      % (name, value))
+
+
 class CustomDimension(BasicDimension):
 
     """
@@ -1072,8 +1147,10 @@ class CustomDimension(BasicDimension):
 
     is_Custom = True
 
+    __rkwargs__ = ('symbolic_min', 'symbolic_max', 'symbolic_size', 'parent')
+
     def __init_finalize__(self, name, symbolic_min=None, symbolic_max=None,
-                          symbolic_size=None, parent=None):
+                          symbolic_size=None, parent=None, **kwargs):
         self._symbolic_min = symbolic_min
         self._symbolic_max = symbolic_max
         self._symbolic_size = symbolic_size
@@ -1102,6 +1179,13 @@ class CustomDimension(BasicDimension):
         else:
             return self._spacing
 
+    @property
+    def bound_symbols(self):
+        if self.is_Derived:
+            return self.parent.bound_symbols
+        else:
+            return frozenset()
+
     @cached_property
     def _defines(self):
         ret = frozenset({self})
@@ -1152,9 +1236,6 @@ class CustomDimension(BasicDimension):
         """A CustomDimension performs no runtime checks."""
         return
 
-    # Pickling support
-    _pickle_kwargs = ['symbolic_min', 'symbolic_max', 'symbolic_size', 'parent']
-
 
 class DynamicDimensionMixin(object):
 
@@ -1187,6 +1268,70 @@ class DynamicSubDimension(DynamicDimensionMixin, SubDimension):
                 Scalar(name="%s_rtkn" % name, dtype=np.int32, nonnegative=True))
 
 
+class StencilDimension(BasicDimension):
+
+    """
+    Dimension symbol representing the points of a stencil.
+
+    Parameters
+    ----------
+    name : str
+        Name of the dimension.
+    _min : expr-like
+        The minimum point of the stencil.
+    _max : expr-like
+        The maximum point of the stencil.
+    spacing : expr-like, optional
+        The space between two stencil points.
+    """
+
+    def __init_finalize__(self, name, _min, _max, spacing=None, **kwargs):
+        self._spacing = sympy.sympify(spacing) or sympy.S.One
+
+        if not is_integer(_min):
+            raise ValueError("Expected integer `min` (got %s)" % _min)
+        if not is_integer(_max):
+            raise ValueError("Expected integer `max` (got %s)" % _max)
+        if not is_integer(self._spacing):
+            raise ValueError("Expected integer `spacing` (got %s)" % self._spacing)
+
+        self._min = _min
+        self._max = _max
+        self._size = _max - _min + 1
+
+        if self._size < 1:
+            raise ValueError("Expected size greater than 0 (got %s)" % self._size)
+
+    def _hashable_content(self):
+        return super()._hashable_content() + (self._min, self._max, self._spacing)
+
+    @cached_property
+    def symbolic_size(self):
+        return sympy.Number(self._size)
+
+    @cached_property
+    def symbolic_min(self):
+        return sympy.Number(self._min)
+
+    @cached_property
+    def symbolic_max(self):
+        return sympy.Number(self._max)
+
+    @property
+    def range(self):
+        return range(self._min, self._max + 1)
+
+    @property
+    def _arg_names(self):
+        return ()
+
+    def _arg_defaults(self, **kwargs):
+        return {}
+
+    def _arg_values(self, *args, **kwargs):
+        return {}
+
+
 # ***
 # The Dimensions below are created by Devito and may eventually be
 # accessed in user code to e.g. construct or manipulate Eqs
@@ -1262,80 +1407,13 @@ class SteppingDimension(DerivedDimension):
         return values
 
 
-class BlockDimension(AbstractIncrDimension):
-
-    """
-    Dimension symbol for lowering TILABLE Dimensions.
-    """
-
-    is_Block = True
-    is_PerfKnob = True
-
-    @cached_property
-    def _arg_names(self):
-        try:
-            return (self.step.name,)
-        except AttributeError:
-            # `step` not a Symbol
-            return ()
-
-    def _arg_defaults(self, **kwargs):
-        # TODO: need a heuristic to pick a default incr size
-        # TODO: move default value to __new__
-        try:
-            return {self.step.name: 8}
-        except AttributeError:
-            # `step` not a Symbol
-            return {}
-
-    def _arg_values(self, interval, grid, args=None, **kwargs):
-        try:
-            name = self.step.name
-        except AttributeError:
-            # `step` not a Symbol
-            return {}
-
-        if name in kwargs:
-            return {name: kwargs.pop(name)}
-        elif isinstance(self.parent, BlockDimension):
-            # `self` is a BlockDimension within an outer BlockDimension, but
-            # no value supplied -> the sub-block will span the entire block
-            return {name: args[self.parent.step.name]}
-        else:
-            value = self._arg_defaults()[name]
-            if value <= args[self.root.max_name] - args[self.root.min_name] + 1:
-                return {name: value}
-            else:
-                # Avoid OOB (will end up here only in case of tiny iteration spaces)
-                return {name: 1}
-
-    def _arg_check(self, args, *_args):
-        try:
-            name = self.step.name
-        except AttributeError:
-            # `step` not a Symbol
-            return
-
-        value = args[name]
-        if isinstance(self.parent, BlockDimension):
-            # sub-BlockDimensions must be perfect divisors of their parent
-            parent_value = args[self.parent.step.name]
-            if parent_value % value > 0:
-                raise InvalidArgument("Illegal block size `%s=%d`: sub-block sizes "
-                                      "must divide the parent block size evenly (`%s=%d`)"
-                                      % (name, value, self.parent.step.name,
-                                         parent_value))
-        else:
-            if value < 0:
-                raise InvalidArgument("Illegal block size `%s=%d`: it should be > 0"
-                                      % (name, value))
-            if value > args[self.root.max_name] - args[self.root.min_name] + 1:
-                # Avoid OOB
-                raise InvalidArgument("Illegal block size `%s=%d`: it's greater than the "
-                                      "iteration range and it will cause an OOB access"
-                                      % (name, value))
+# ***
+# Utils
 
 
 def dimensions(names):
     assert type(names) == str
     return tuple(Dimension(i) for i in names.split())
+
+
+BOTTOM = Dimension(name='⊥')
